import o3seespy as o3


def get_o3_class_and_args_from_soil_obj(sl, saturated=False, esig_v0=None, f_order=1e3, overrides=None):
    if overrides is None:
        overrides = {}

    if saturated:
        umass = sl.unit_sat_mass / f_order  # TODO: work out how to run in Pa, N, m, s
    else:
        umass = sl.unit_dry_mass / f_order
    if 'p_atm' not in overrides:
        if not hasattr(sl, 'p_atm'):
            p_atm = 101.0e3
        else:
            p_atm = sl.p_atm
        overrides['p_atm'] = p_atm / f_order
    if 'rho' not in overrides:
        overrides['rho'] = umass
    if 'unit_moist_mass' not in overrides:
        overrides['unit_moist_mass'] = umass
    if 'nd' not in overrides:
        overrides['nd'] = 2

    app2mod = {}
    # Define material
    if not hasattr(sl, 'o3_type'):
        sl.o3_type = sl.type  # for backward compatibility
    if sl.o3_type == 'pm4sand':
        sl_class = o3.nd_material.PM4Sand
        app2mod = sl.app2mod
    elif sl.o3_type == 'sdmodel':
        sl_class = o3.nd_material.StressDensity
        # overrides = {'nu': pois, 'p_atm': 101, 'unit_moist_mass': umass}
        app2mod = sl.app2mod
    elif sl.o3_type in ['pimy', 'pdmy', 'pdmy02']:
        if hasattr(sl, 'get_g_mod_at_m_eff_stress'):
            if hasattr(sl, 'g_mod_p0') and sl.g_mod_p0 != 0.0:
                k0 = sl.poissons_ratio / (1 - sl.poissons_ratio)
                m_eff = esig_v0 * (1 + 2 * k0) / 3
                p = m_eff  # Pa
                overrides['d'] = 0.0
            else:
                p = 101.0e3  # Pa
                overrides['d'] = sl.a
            g_mod_r = sl.get_g_mod_at_m_eff_stress(p)
        else:
            p = 101.0e3  # Pa
            overrides['d'] = 0.0
            g_mod_r = sl.g_mod

        b_mod = 2 * g_mod_r * (1 + sl.poissons_ratio) / (3 * (1 - 2 * sl.poissons_ratio))
        overrides['p_ref'] = p / f_order
        overrides['g_mod_ref'] = g_mod_r / f_order
        overrides['bulk_mod_ref'] = b_mod / f_order
        if sl.o3_type == 'pimy':
            overrides['cohesion'] = sl.cohesion / f_order
            sl_class = o3.nd_material.PressureIndependMultiYield
        elif sl.o3_type == 'pdmy':
            sl_class = o3.nd_material.PressureDependMultiYield
        elif sl.o3_type == 'pdmy02':
            sl_class = o3.nd_material.PressureDependMultiYield02
    else:
        if hasattr(sl, 'get_g_mod_at_v_eff_stress'):
            g_mod = sl.get_g_mod_at_v_eff_stress(esig_v0)
        else:
            g_mod = sl.g_mod
        sl_class = o3.nd_material.ElasticIsotropic
        sl.e_mod = 2 * g_mod * (1 - sl.poissons_ratio)
        overrides['nu'] = sl.poissons_ratio
        app2mod['rho'] = 'unit_moist_mass'
    args, kwargs = o3.extensions.get_o3_kwargs_from_obj(sl, sl_class, custom=app2mod, overrides=overrides)
    return sl_class, args, kwargs



import o3seespy as o3
import numpy as np


def run_vload(mat, v_pressure, osi=None, nu_dyn=None):
    if osi is None:
        osi = o3.OpenSeesInstance(ndm=2, ndf=2)
        mat.build(osi)
    h_ele = 1.
    nodes = [
        o3.node.Node(osi, 0.0, 0.0),
        o3.node.Node(osi, h_ele, 0.0),
        o3.node.Node(osi, h_ele, h_ele),
        o3.node.Node(osi, 0.0, h_ele)
    ]

    # Fix bottom node
    o3.Fix2DOF(osi, nodes[0], o3.cc.FIXED, o3.cc.FIXED)
    o3.Fix2DOF(osi, nodes[1], o3.cc.FIXED, o3.cc.FIXED)
    o3.Fix2DOF(osi, nodes[2], o3.cc.FREE, o3.cc.FREE)
    o3.Fix2DOF(osi, nodes[3], o3.cc.FREE, o3.cc.FREE)
    # Set out-of-plane DOFs to be slaved
    o3.EqualDOF(osi, nodes[2], nodes[3], [o3.cc.X, o3.cc.Y])

    ele = o3.element.SSPquad(osi, nodes, mat, 'PlaneStrain', 1, 0.0, 0.0)

    o3.constraints.Transformation(osi)
    o3.test_check.NormDispIncr(osi, tol=1.0e-3, max_iter=35, p_flag=0)
    o3.algorithm.Newton(osi)
    o3.numberer.RCM(osi)
    o3.system.FullGeneral(osi)
    # o3.integrator.DisplacementControl(osi, nodes[2], o3.cc.DOF2D_Y, 0.005)
    # o3.integrator.Newmark(osi, gamma=5. / 6, beta=4. / 9)
    o3.integrator.LoadControl(osi, 1)

    # o3.rayleigh.Rayleigh(osi, a0, a1, 0.0, 0.0)
    o3.analysis.Static(osi)
    if hasattr(mat, 'update_to_nonlinear'):
        mat.update_to_nonlinear()
    if nu_dyn is not None:
        mat.set_poissons_ratio(osi, nu_dyn, ele=ele)

    ods = {'stresses': o3.recorder.ElementToArrayCache(osi, ele=ele, arg_vals=['stress']),
           'strains': o3.recorder.ElementToArrayCache(osi, ele=ele, arg_vals=['strain'])
           }

    ts0 = o3.time_series.Path(osi, time=[0, 1000, 1e10],
                              values=[0, 1, 1], factor=1)
    pat0 = o3.pattern.Plain(osi, ts0)
    o3.Load(osi, nodes[2], [0, -v_pressure * h_ele / 2])
    o3.Load(osi, nodes[3], [0, -v_pressure * h_ele / 2])
    stresses = o3.get_ele_response(osi, ele, 'stress')
    o3.record(osi)
    for i in range(1000):
        o3.analyze(osi, num_inc=1)
        stresses = o3.get_ele_response(osi, ele, 'stress')
        if stresses[1] >= v_pressure:
            break
    o3.wipe(osi)
    for item in ods:
        ods[item] = ods[item].collect()
    return ods['stresses'], ods['strains']


def run_example(show=0):
    osi = o3.OpenSeesInstance(ndm=2, ndf=2, state=3)

    esig_v0 = 50.0e3
    poissons_ratio = 0.3
    g_mod = 1.0e6
    b_mod = 2 * g_mod * (1 + poissons_ratio) / (3 * (1 - 2 * poissons_ratio))

    mat = o3.nd_material.PressureIndependMultiYield(osi, 2, 2058.49, g_mod, b_mod, 68000.0, 0.1, 0.0, 100000.0, 0.0, 25)
    # mat = o3.nd_material.ElasticIsotropic(osi, e_mod=1.0e6, nu=0.3)
    ss, es = run_vload(mat, v_pressure=esig_v0, osi=osi)
    if show:
        import matplotlib.pyplot as plt
        plt.plot(es[:, 1], ss[:, 1])
        plt.show()


if __name__ == '__main__':
    run_example(show=1)



import numpy as np
import sfsimodels as sm
import openseespy.opensees as opy
import o3seespy as o3
import o3seespy.extensions
import copy
import os


def site_response(sp, asig, freqs=(0.5, 10), xi=0.03, analysis_dt=0.001, dy=0.5, analysis_time=None, outs=None,
                  rec_dt=None, base_imp=0, cache_path=None, pload=0.0, opfile=None):
    """
    Run seismic analysis of a soil profile - example based on:
    http://opensees.berkeley.edu/wiki/index.php/Site_Response_Analysis_of_a_Layered_Soil_Column_(Total_Stress_Analysis)

    Parameters
    ----------
    sp: sfsimodels.SoilProfile object
        A soil profile
    asig: eqsig.AccSignal object
        An acceleration signal
    base_imp: float
        If positive then use as impedence at base of model,
        If zero then use last soil layer
        If negative then use fixed base

    Returns
    -------

    """
    if analysis_time is None:
        analysis_time = asig.time[-1]
    if outs is None:
        outs = {'ACCX': [0]}  # Export the horizontal acceleration at the surface
    if rec_dt is None:
        rec_dt = analysis_dt
    else:
        raise ValueError('This is causing an error')
    state = 0
    if opfile:
        state = 3
    osi = o3.OpenSeesInstance(ndm=2, ndf=2, state=state)
    assert isinstance(sp, sm.SoilProfile)
    sp.gen_split(props=['shear_vel', 'unit_mass'], target=dy)
    thicknesses = sp.split["thickness"]
    n_node_rows = len(thicknesses) + 1
    node_depths = np.cumsum(sp.split["thickness"])
    node_depths = np.insert(node_depths, 0, 0)
    ele_depths = (node_depths[1:] + node_depths[:-1]) / 2
    unit_masses = sp.split["unit_mass"] / 1e3

    grav = 9.81
    # Rayleigh damping parameters
    omega_1 = 2 * np.pi * freqs[0]
    omega_2 = 2 * np.pi * freqs[1]
    a0 = 2 * xi * omega_1 * omega_2 / (omega_1 + omega_2)
    a1 = 2 * xi / (omega_1 + omega_2)

    k0 = 0.5
    pois = k0 / (1 + k0)

    ele_width = 3 * min(thicknesses)

    # Define nodes and set boundary conditions for simple shear deformation
    # Start at top and build down?
    nx = 1
    sn = []
    # sn = [[o3.node.Node(osi, ele_width * j, 0) for j in range(nx + 1)]]
    for i in range(0, n_node_rows):
        # Establish left and right nodes
        sn.append([o3.node.Node(osi, ele_width * j, -node_depths[i]) for j in range(nx + 1)])
        # set x and y dofs equal for left and right nodes
        o3.EqualDOF(osi, sn[i][0], sn[i][-1], [o3.cc.X, o3.cc.Y])
    sn = np.array(sn)

    if base_imp < 0:
        # Fix base nodes
        for j in range(nx + 1):
            o3.Fix2DOF(osi, sn[-1][j], o3.cc.FIXED, o3.cc.FIXED)
    else:
        # Fix base nodes
        for j in range(nx + 1):
            o3.Fix2DOF(osi, sn[-1][j], o3.cc.FREE, o3.cc.FIXED)

        # Define dashpot nodes
        dashpot_node_l = o3.node.Node(osi, 0, -node_depths[-1])
        dashpot_node_2 = o3.node.Node(osi, 0, -node_depths[-1])
        o3.Fix2DOF(osi, dashpot_node_l, o3.cc.FIXED, o3.cc.FIXED)
        o3.Fix2DOF(osi, dashpot_node_2, o3.cc.FREE, o3.cc.FIXED)

        # define equal DOF for dashpot and soil base nodes
        o3.EqualDOF(osi, sn[-1][0], sn[-1][1], [o3.cc.X])
        o3.EqualDOF(osi, sn[-1][0], dashpot_node_2, [o3.cc.X])

    # define materials
    ele_thick = 1.0  # m
    soil_mats = []
    strains = np.logspace(-6, -0.5, 16)
    prev_args = []
    prev_kwargs = {}
    prev_sl_type = None
    eles = []
    for i in range(len(thicknesses)):
        y_depth = ele_depths[i]

        sl_id = sp.get_layer_index_by_depth(y_depth)
        sl = sp.layer(sl_id)
        app2mod = {}
        if y_depth > sp.gwl:
            umass = sl.unit_sat_mass / 1e3
        else:
            umass = sl.unit_dry_mass / 1e3
        # Define material
        if sl.type == 'pm4sand':
            sl_class = o3.nd_material.PM4Sand
            overrides = {'nu': pois, 'p_atm': 101, 'unit_moist_mass': umass}
            app2mod = sl.app2mod
        elif sl.type == 'sdmodel':
            sl_class = o3.nd_material.StressDensity
            overrides = {'nu': pois, 'p_atm': 101, 'unit_moist_mass': umass}
            app2mod = sl.app2mod
        elif sl.type == 'pimy':
            sl_class = o3.nd_material.PressureIndependMultiYield
            overrides = {'nu': pois, 'p_atm': 101,
                         'rho': umass,
                         'nd': 2.0,
                         'g_mod_ref': sl.g_mod / 1e3,
                         'bulk_mod_ref': sl.bulk_mod / 1e3,
                         'cohesion': sl.cohesion / 1e3,
                         'p_ref': sl.p_ref / 1e3,
                         'd': 0.0
                         # 'n_surf': 25
                         }
        else:
            sl_class = o3.nd_material.ElasticIsotropic
            sl.e_mod = 2 * sl.g_mod * (1 - sl.poissons_ratio) / 1e3
            app2mod['rho'] = 'unit_moist_mass'
            overrides = {'nu': sl.poissons_ratio, 'unit_moist_mass': umass}

        args, kwargs = o3.extensions.get_o3_kwargs_from_obj(sl, sl_class, custom=app2mod, overrides=overrides)
        changed = 0
        if sl.type != prev_sl_type or len(args) != len(prev_args) or len(kwargs) != len(prev_kwargs):
            changed = 1
        else:
            for j, arg in enumerate(args):
                if not np.isclose(arg, prev_args[j]):
                    changed = 1
            for pm in kwargs:
                if pm not in prev_kwargs or not np.isclose(kwargs[pm], prev_kwargs[pm]):
                    changed = 1

        if changed:
            mat = sl_class(osi, *args, **kwargs)
            prev_sl_type = sl.type
            prev_args = copy.deepcopy(args)
            prev_kwargs = copy.deepcopy(kwargs)

            soil_mats.append(mat)

        # def element
        for xx in range(nx):
            nodes = [sn[i+1][xx], sn[i+1][xx + 1], sn[i][xx + 1], sn[i][xx]]  # anti-clockwise
            # ele = o3.element.Quad(osi, nodes, ele_thick, o3.cc.PLANE_STRAIN, mat, b2=grav * unit_masses[i])
            # eles.append(ele)
            eles.append(o3.element.SSPquad(osi, nodes, mat, o3.cc.PLANE_STRAIN, ele_thick, 0.0, grav * unit_masses[i]))

    if base_imp >= 0:
        # define material and element for viscous dampers
        base_sl = sp.layer(sp.n_layers)
        c_base = ele_width * base_sl.unit_dry_mass / 1e3 * sp.get_shear_vel_at_depth(sp.height)
        dashpot_mat = o3.uniaxial_material.Viscous(osi, c_base, alpha=1.)
        o3.element.ZeroLength(osi, [dashpot_node_l, dashpot_node_2], mats=[dashpot_mat], dirs=[o3.cc.DOF2D_X])

    # Static analysis
    o3.constraints.Transformation(osi)
    o3.test.NormDispIncr(osi, tol=1.0e-5, max_iter=30, p_flag=0)
    o3.algorithm.Newton(osi)
    o3.numberer.RCM(osi)
    o3.system.ProfileSPD(osi)
    o3.integrator.Newmark(osi, gamma=0.5, beta=0.25)
    o3.analysis.Transient(osi)
    o3.analyze(osi, 10, 500.)

    for i in range(len(soil_mats)):
        if hasattr(soil_mats[i], 'update_to_nonlinear'):
            soil_mats[i].update_to_nonlinear(osi)
    o3.analyze(osi, 40, 500.)

    o3.rayleigh.Rayleigh(osi, a0, a1, 0, 0)

    # reset time and analysis
    o3.set_time(osi, 0.0)  # TODO:
    o3.wipe_analysis(osi)

    coords = o3.get_all_node_coords(osi)
    ele_node_tags = o3.get_all_ele_node_tags_as_dict(osi)
    all_node_xdisp_rec = o3.recorder.NodesToArrayCache(osi, 'all', [o3.cc.DOF2D_X], 'disp', nsd=4)
    all_node_ydisp_rec = o3.recorder.NodesToArrayCache(osi, 'all', [o3.cc.DOF2D_Y], 'disp', nsd=4)

    # Define the dynamic analysis
    print('Here')
    o3.constraints.Transformation(osi)
    o3.test.NormDispIncr(osi, tol=1.0e-5, max_iter=15, p_flag=0)
    #o3.test_check.EnergyIncr(osi, tol=1.0e-7, max_iter=10)
    o3.algorithm.Newton(osi)
    o3.system.SparseGeneral(osi)
    o3.numberer.RCM(osi)
    o3.integrator.Newmark(osi, gamma=0.5, beta=0.25)
    o3.analysis.Transient(osi)

    if pload:
        static_time = 100
        print('time: ', o3.get_time(osi))
        # Add static stress bias
        time_series = o3.time_series.Path(osi, time=[0, static_time / 2, static_time, 1e3], values=[0, 0.5, 1, 1], use_last=1)
        o3.pattern.Plain(osi, time_series)
        o3.Load(osi, sn[0][0], [pload * ele_width, 0])
        o3.Load(osi, sn[9][0], [-pload * ele_width, 0])
        if base_imp >= 0:
            o3.Load(osi, sn[-1][0], [-pload, 0])

        static_dt = 0.1
        o3.analyze(osi, int(static_time / static_dt) * 1.5, static_dt)
        o3.load_constant(osi, time=0)

    o3.set_time(osi, 0.0)  # TODO:
    init_time = o3.get_time(osi)
    o3sra_outs = O3SRAOutputs()
    o3sra_outs.start_recorders(osi, outs, sn, eles, rec_dt=rec_dt)

    # Define the dynamic input motion
    if base_imp < 0:  # fixed base
        acc_series = o3.time_series.Path(osi, dt=asig.dt, values=-asig.values)  # should be negative
        o3.pattern.UniformExcitation(osi, dir=o3.cc.X, accel_series=acc_series)
    else:
        ts_obj = o3.time_series.Path(osi, dt=asig.dt, values=asig.velocity * -1, factor=c_base)
        o3.pattern.Plain(osi, ts_obj)
        o3.Load(osi, sn[-1][0], [1., 0.])

    if state == 3:
        o3.extensions.to_py_file(osi, opfile)
    # Run the dynamic motion
    while o3.get_time(osi) - init_time < analysis_time:
        if o3.analyze(osi, 1, analysis_dt):
            print('failed')
            break
    o3.wipe(osi)
    out_dict = o3sra_outs.results_to_dict()

    if cache_path:
        import o3_plot
        o3sra_outs.cache_path = cache_path
        o3sra_outs.results_to_files()
        o3res = o3_plot.O3Results()
        o3res.cache_path = cache_path
        o3res.coords = coords
        o3res.ele_node_tags = ele_node_tags
        o3res.x_disp = all_node_xdisp_rec.collect()
        o3res.y_disp = all_node_ydisp_rec.collect()
        o3res.save_to_cache()

    return out_dict



class O3SRAOutputs(object):
    cache_path = ''
    out_dict = None
    area = 1.0
    outs = None

    def start_recorders(self, osi, outs, sn, eles, rec_dt, sn_xy=False):
        self.rec_dt = rec_dt
        self.eles = eles
        self.sn_xy = sn_xy
        if sn_xy:
            self.nodes = sn[0, :]
        else:
            self.nodes = sn[:, 0]
        self.outs = outs
        node_depths = np.array([node.y for node in sn[:, 0]])
        ele_depths = (node_depths[1:] + node_depths[:-1]) / 2
        ods = {}
        for otype in outs:
            if otype in ['ACCX', 'DISPX']:
                if isinstance(outs[otype], str) and outs[otype] == 'all':

                    if otype == 'ACCX':
                        ods['ACCX'] = o3.recorder.NodesToArrayCache(osi, nodes=self.nodes, dofs=[o3.cc.X], res_type='accel',
                                                                dt=rec_dt)
                    if otype == 'DISPX':
                        ods['DISPX'] = o3.recorder.NodesToArrayCache(osi, nodes=self.nodes, dofs=[o3.cc.X], res_type='disp',
                                                                dt=rec_dt)
                else:
                    ods['ACCX'] = []
                    for i in range(len(outs['ACCX'])):
                        ind = np.argmin(abs(node_depths - outs['ACCX'][i]))
                        ods['ACCX'].append(
                            o3.recorder.NodeToArrayCache(osi, node=sn[ind][0], dofs=[o3.cc.X], res_type='accel', dt=rec_dt))
            if otype == 'TAU':
                for ele in eles:
                    assert isinstance(ele, o3.element.SSPquad)
                ods['TAU'] = []
                if isinstance(outs['TAU'], str) and outs['TAU'] == 'all':
                    ods['TAU'] = o3.recorder.ElementsToArrayCache(osi, eles=eles, arg_vals=['stress'], dt=rec_dt)
                else:
                    for i in range(len(outs['TAU'])):
                        ind = np.argmin(abs(ele_depths - outs['TAU'][i]))
                        ods['TAU'].append(
                            o3.recorder.ElementToArrayCache(osi, ele=eles[ind], arg_vals=['stress'], dt=rec_dt))
            if otype == 'TAUX':
                if isinstance(outs['TAUX'], str) and outs['TAUX'] == 'all':
                    if sn_xy:
                        order = 'F'
                    else:
                        order = 'C'
                    ods['TAUX'] = o3.recorder.NodesToArrayCache(osi, nodes=sn.flatten(order), dofs=[o3.cc.X], res_type='reaction',
                                                                dt=rec_dt)
            if otype == 'STRS':
                ods['STRS'] = []
                if isinstance(outs['STRS'], str) and outs['STRS'] == 'all':
                    ods['STRS'] = o3.recorder.ElementsToArrayCache(osi, eles=eles, arg_vals=['strain'], dt=rec_dt)
                else:
                    for i in range(len(outs['STRS'])):
                        ind = np.argmin(abs(ele_depths - outs['STRS'][i]))
                        ods['STRS'].append(o3.recorder.ElementToArrayCache(osi, ele=eles[ind], arg_vals=['strain'], dt=rec_dt))
            if otype == 'STRSX':
                if isinstance(outs['STRSX'], str) and outs['STRSX'] == 'all':
                    if 'DISPX' in outs:
                        continue
                    if sn_xy:
                        nodes = sn[0, :]
                    else:
                        nodes = sn[:, 0]
                    ods['DISPX'] = o3.recorder.NodesToArrayCache(osi, nodes=nodes, dofs=[o3.cc.X], res_type='disp',
                                                                dt=rec_dt)

        self.ods = ods

    def results_to_files(self):
        od = self.results_to_dict()
        for item in od:
            ffp = self.cache_path + f'{item}.txt'
            if os.path.exists(ffp):
                os.remove(ffp)
            np.savetxt(ffp, od[item])

    def load_results_from_files(self, outs=None):
        if outs is None:
            outs = ['ACCX', 'TAU', 'STRS', 'time']
        od = {}
        for item in outs:
            od[item] = np.loadtxt(self.cache_path + f'{item}.txt')
        return od

    def results_to_dict(self):
        ro = o3.recorder.load_recorder_options()
        import pandas as pd
        df = pd.read_csv(ro)
        if self.outs is None:
            items = list(self.ods)
        else:
            items = list(self.outs)
        if self.out_dict is None:
            self.out_dict = {}
            for otype in items:
                if otype not in self.ods:
                    if otype == 'STRSX':
                        depths = []
                        for node in self.nodes:
                            depths.append(node.y)
                        depths = np.array(depths)
                        d_incs = depths[1:] - depths[:-1]
                        vals = self.ods['DISPX'].collect(unlink=False).T
                        self.out_dict[otype] = (vals[1:] - vals[:-1]) / d_incs[:, np.newaxis]
                elif isinstance(self.ods[otype], list):
                    self.out_dict[otype] = []
                    for i in range(len(self.ods[otype])):
                        if otype in ['TAU', 'STRS']:
                            self.out_dict[otype].append(self.ods[otype][i].collect()[2])
                        else:
                            self.out_dict[otype].append(self.ods[otype][i].collect())
                    self.out_dict[otype] = np.array(self.out_dict[otype])
                else:
                    vals = self.ods[otype].collect().T
                    cur_ind = 0
                    self.out_dict[otype] = []
                    if otype in ['TAU', 'STRS']:
                        for ele in self.eles:
                            mat_type = ele.mat.type
                            form = 'PlaneStrain'
                            dfe = df[(df['mat'] == mat_type) & (df['form'] == form)]
                            if otype == 'TAU':
                                dfe = dfe[dfe['recorder'] == 'stress']
                                ostr = 'sxy'
                            else:
                                dfe = dfe[dfe['recorder'] == 'strain']
                                ostr = 'gxy'
                            assert len(dfe) == 1
                            outs = dfe['outs'].iloc[0].split('-')
                            oind = outs.index(ostr)
                            self.out_dict[otype].append(vals[cur_ind + oind])
                            cur_ind += len(outs)
                        self.out_dict[otype] = np.array(self.out_dict[otype])
                        # if otype == 'STRS':
                        #     self.out_dict[otype] = vals[2::3]  # Assumes pimy
                        # elif otype == 'TAU':
                        #     self.out_dict[otype] = vals[3::5]  # Assumes pimy
                    elif otype == 'TAUX':
                        f_static = -np.cumsum(vals[::2, :] - vals[1::2, :], axis=0)[:-1]  # add left and right
                        f_dyn = vals[::2, :] + vals[1::2, :]  # add left and right
                        f_dyn_av = (f_dyn[1:] + f_dyn[:-1]) / 2
                        # self.out_dict[otype] = (f[1:, :] - f[:-1, :]) / area
                        self.out_dict[otype] = (f_dyn_av + f_static) / self.area
                    else:
                        self.out_dict[otype] = vals
            # Create time output
            if 'ACCX' in self.out_dict:
                self.out_dict['time'] = np.arange(0, len(self.out_dict['ACCX'][0])) * self.rec_dt
            elif 'TAU' in self.out_dict:
                self.out_dict['time'] = np.arange(0, len(self.out_dict['TAU'][0])) * self.rec_dt
        return self.out_dict


def site_response_w_pysra(soil_profile, asig, odepths):
    import liquepy as lq
    import pysra
    pysra_profile = lq.sra.sm_profile_to_pysra(soil_profile, d_inc=[0.5] * soil_profile.n_layers)
    # Should be input in g
    pysra_m = pysra.motion.TimeSeriesMotion(asig.label, None, time_step=asig.dt, accels=-asig.values / 9.8)

    calc = pysra.propagation.EquivalentLinearCalculator()

    od = {'ACCX': [], 'STRS': [], 'TAU': []}
    outs = []
    for i, depth in enumerate(odepths):
        od['ACCX'].append(len(outs))
        outs.append(pysra.output.AccelerationTSOutput(pysra.output.OutputLocation('within', depth=depth)))
        od['STRS'].append(len(outs))
        outs.append(pysra.output.StrainTSOutput(pysra.output.OutputLocation('within', depth=depth), in_percent=False))
        od['TAU'].append(len(outs))
        outs.append(pysra.output.StressTSOutput(pysra.output.OutputLocation('within', depth=depth),
                                                normalized=False))
    outputs = pysra.output.OutputCollection(outs)
    calc(pysra_m, pysra_profile, pysra_profile.location('outcrop', depth=soil_profile.height))
    outputs(calc)

    out_series = {}
    for mtype in od:
        out_series[mtype] = []
        for i in range(len(od[mtype])):
            out_series[mtype].append(outputs[od[mtype][i]].values[:asig.npts])
        out_series[mtype] = np.array(out_series[mtype])
        if mtype == 'ACCX':
            out_series[mtype] *= 9.8
    return out_series



import o3seespy as o3
import math


def run_ud_cdss(mat, esig_v0, csr, osi=None, static_bias=0.0, n_lim=100, nu_dyn=None, opyfile=None,
                strain_limit=0.03, strain_inc=5.0e-6, verbose=0):
    """Undrained cyclic simple shear test for 2d element"""
    damp = 0.02
    omega0 = 0.2
    omega1 = 20.0
    a1 = 2. * damp / (omega0 + omega1)
    a0 = a1 * omega0 * omega1

    if osi is None:
        osi = o3.OpenSeesInstance(ndm=2, ndf=2)
        mat.build(osi)

    # Establish nodes
    h_ele = 1.
    bl_node = o3.node.Node(osi, 0, 0)
    br_node = o3.node.Node(osi, h_ele, 0)
    tr_node = o3.node.Node(osi, h_ele, h_ele)
    tl_node = o3.node.Node(osi, 0, h_ele)
    all_nodes = [bl_node, br_node, tr_node, tl_node]

    # Fix bottom node
    o3.Fix3DOF(osi, bl_node, o3.cc.FIXED, o3.cc.FIXED, o3.cc.FIXED)
    o3.Fix3DOF(osi, br_node, o3.cc.FIXED, o3.cc.FIXED, o3.cc.FIXED)
    o3.Fix3DOF(osi, tr_node, o3.cc.FREE, o3.cc.FREE, o3.cc.FIXED)
    o3.Fix3DOF(osi, tl_node, o3.cc.FREE, o3.cc.FREE, o3.cc.FIXED)
    # Set out-of-plane DOFs to be slaved
    o3.EqualDOF(osi, tr_node, tl_node, [o3.cc.X, o3.cc.Y])

    # Note water bulk modulus, permeability, void ratio are irrelevant, since constant volume test
    # - so as soil skeleton contracts
    # the bulk modulus of the soil skeleton controls the change in effective stress
    water_bulk_mod = 2.2e6
    ele = o3.element.SSPquadUP(osi, all_nodes, mat, 1.0, water_bulk_mod, 1.,
                                1.0e-4, 1.0e-4, 0.6, alpha=1.0e-5, b1=0.0, b2=0.0)

    o3.constraints.Transformation(osi)
    o3.test_check.NormDispIncr(osi, tol=1.0e-6, max_iter=35, p_flag=0)
    o3.algorithm.Newton(osi)
    o3.numberer.RCM(osi)
    o3.system.FullGeneral(osi)
    o3.integrator.Newmark(osi, gamma=5./6, beta=4./9)
    o3.rayleigh.Rayleigh(osi, a0, a1, 0.0, 0.0)
    o3.analysis.Transient(osi)

    o3.update_material_stage(osi, mat, stage=0)
    # print('here1: ', o3.get_ele_response(osi, ele, 'stress'), esig_v0, csr)

    all_stresses_cache = o3.recorder.ElementToArrayCache(osi, ele, arg_vals=['stress'])
    all_strains_cache = o3.recorder.ElementToArrayCache(osi, ele, arg_vals=['strain'])
    nodes_cache = o3.recorder.NodesToArrayCache(osi, all_nodes, dofs=[1, 2, 3], res_type='disp')
    o3.recorder.NodesToFile(osi, 'node_disp.txt', all_nodes, dofs=[1, 2, 3], res_type='disp')

    # Add static vertical pressure and stress bias
    time_series = o3.time_series.Path(osi, time=[0, 100, 1e10], values=[0, 1, 1])
    o3.pattern.Plain(osi, time_series)
    o3.Load(osi, tl_node, [0, -esig_v0 / 2, 0])
    o3.Load(osi, tr_node, [0, -esig_v0 / 2, 0])

    o3.analyze(osi, num_inc=110, dt=1)

    ts2 = o3.time_series.Path(osi, time=[110, 80000, 1e10], values=[1., 1., 1.], factor=1)
    o3.pattern.Plain(osi, ts2, fact=1.)
    y_vert = o3.get_node_disp(osi, tr_node, o3.cc.Y)
    o3.SP(osi, tl_node, dof=o3.cc.Y, dof_values=[y_vert])
    o3.SP(osi, tr_node, dof=o3.cc.Y, dof_values=[y_vert])

    # Close the drainage valves
    for node in all_nodes:
        o3.remove_sp(osi, node, dof=3)
    o3.analyze(osi, 25, dt=1)
    if verbose:
        print('Finished applying vert load: ', o3.get_ele_response(osi, ele, 'stress'))

    if hasattr(mat, 'update_to_nonlinear'):
        mat.update_to_nonlinear()
        o3.analyze(osi, 25, dt=1)
    if hasattr(mat, 'set_first_call'):
        mat.set_first_call(value=0, ele=ele)
    # o3.set_parameter(osi, value=0, eles=[ele], args=['FirstCall', mat.tag])
    o3.analyze(osi, 25, dt=1)
    if nu_dyn is not None:
        mat.set_nu(nu_dyn, ele=ele)

    ro = o3.recorder.load_recorder_options()
    import pandas as pd
    df = pd.read_csv(ro)
    mat_type = ele.mat.type
    oop = o3.cc.PLANE_STRAIN
    dfe = df[(df['mat'] == mat_type) & (df['form'] == oop)]
    df_sxy = dfe[dfe['recorder'] == 'stress']
    outs = df_sxy['outs'].iloc[0].split('-')
    sxy_ind = outs.index('sxy')

    n_cyc = 0.0
    target_strain = 1.1 * strain_limit
    target_disp = target_strain * h_ele
    limit_reached = 0
    o3.record(osi)
    while n_cyc < n_lim:
        if verbose:
            print('n_cyc: ', n_cyc)
        h_disp = o3.get_node_disp(osi, tr_node, o3.cc.X)
        curr_time = o3.get_time(osi)
        steps = target_strain / strain_inc
        ts0 = o3.time_series.Path(osi, time=[curr_time, curr_time + steps, 1e10], values=[h_disp, target_disp, target_disp], factor=1)
        pat0 = o3.pattern.Plain(osi, ts0)
        o3.SP(osi, tr_node, dof=o3.cc.X, dof_values=[1.0])
        curr_stress = o3.get_ele_response(osi, ele, 'stress')[sxy_ind]
        if math.isnan(curr_stress):
            raise ValueError

        if opyfile:
            o3.extensions.to_py_file(osi, opyfile)
            opyfile = None
        while curr_stress < (csr - static_bias) * esig_v0:
            o3.analyze(osi, 1, dt=1)
            curr_stress = o3.get_ele_response(osi, ele, 'stress')[sxy_ind]
            h_disp = o3.get_node_disp(osi, tr_node, o3.cc.X)

            if h_disp >= target_disp:
                if verbose:
                    print('STRAIN LIMIT REACHED - on load')
                limit_reached = 1
                break
        if limit_reached:
            break
        n_cyc += 0.25
        if verbose:
            print('load reversal, n_cyc: ', n_cyc)
        curr_time = o3.get_time(osi)
        o3.remove_load_pattern(osi, pat0)
        o3.remove(osi, ts0)
        o3.remove_sp(osi, tr_node, dof=o3.cc.X)
        # Reverse cycle
        steps = (h_disp + target_disp) / (strain_inc * h_ele)
        ts0 = o3.time_series.Path(osi, time=[curr_time, curr_time + steps, 1e10],
                                   values=[h_disp, -target_disp, -target_disp], factor=1)
        pat0 = o3.pattern.Plain(osi, ts0)
        o3.SP(osi, tr_node, dof=o3.cc.X, dof_values=[1.0])
        i = 0
        while curr_stress > -(csr + static_bias) * esig_v0:
            o3.analyze(osi, 1, dt=1)
            curr_stress = o3.get_ele_response(osi, ele, 'stress')[2]
            h_disp = o3.get_node_disp(osi, tr_node, o3.cc.X)

            if -h_disp >= target_disp:
                if verbose:
                    print('STRAIN LIMIT REACHED - on reverse')
                limit_reached = 1
                break
            i += 1
            if i > steps:
                break
        if limit_reached:
            break
        n_cyc += 0.5
        if verbose:
            print('reload, n_cyc: ', n_cyc)
        curr_time = o3.get_time(osi)
        o3.remove_load_pattern(osi, pat0)
        o3.remove(osi, ts0)
        o3.remove_sp(osi, tr_node, dof=o3.cc.X)
        # reload cycle
        steps = (-h_disp + target_disp) / (strain_inc * h_ele)
        ts0 = o3.time_series.Path(osi, time=[curr_time, curr_time + steps, 1e10],
                                   values=[h_disp, target_disp, target_disp], factor=1)
        pat0 = o3.pattern.Plain(osi, ts0)
        o3.SP(osi, tr_node, dof=o3.cc.X, dof_values=[1.0])
        while curr_stress < static_bias * esig_v0:
            o3.analyze(osi, 1, dt=1)
            curr_stress = o3.get_ele_response(osi, ele, 'stress')[sxy_ind]
            h_disp = o3.get_node_disp(osi, tr_node, o3.cc.X)

            if h_disp >= target_disp:
                if verbose:
                    print('STRAIN LIMIT REACHED - on reload')
                limit_reached = 1
                break
        if limit_reached:
            break
        o3.remove_load_pattern(osi, pat0)
        o3.remove(osi, ts0)
        o3.remove_sp(osi, tr_node, dof=o3.cc.X)
        n_cyc += 0.25

    o3.wipe(osi)
    all_stresses = all_stresses_cache.collect()
    all_strains = all_strains_cache.collect()
    disps = nodes_cache.collect()
    stress = all_stresses[:, 2]
    strain = all_strains[:, 2]
    ppt = all_stresses[:, 1]

    return stress, strain, ppt, disps


def run_example(show=0):
    import o3seespy as o3

    esig_v0 = 101.3

    # PM4Sand properties
    relative_density = 0.35
    g0_mod = 476.0
    h_po = 0.53
    unit_sat_mass = 1.42
    k0 = 0.5
    nu_dyn = 0.3
    p_atm = 101.0e3
    strain_inc = 5.e-6
    csr = 0.16

    # Initialise OpenSees instance
    osi = o3.OpenSeesInstance(ndm=2, ndf=3, state=3)

    # Define material
    nu_init = k0 / (1 + k0)
    mat = o3.nd_material.PM4Sand(osi, relative_density, g0_mod, h_po, unit_sat_mass, p_atm, nu=nu_init)
    # mat = o3.nd_material.ElasticIsotropic(osi, e_mod=1.0e10, nu=0.3)  # TODO: not working with the elastic model!!!
    # nu_dyn = None
    stress, strain, ppt, disps = run_ud_cdss(mat, csr=csr, osi=osi, n_lim=20, strain_limit=0.03, nu_dyn=nu_dyn,
                                                esig_v0=esig_v0, strain_inc=strain_inc, opyfile='ss.py', verbose=0)

    if show:
        import matplotlib.pyplot as plt
        bf, sps = plt.subplots(nrows=2)
        sps[0].plot(stress, label='shear')
        sps[0].plot(ppt, label='PPT')
        sps[1].plot(strain, stress, label='o3seespy')

        sps[0].set_xlabel('Time [s]')
        sps[0].set_ylabel('Stress [kPa]')
        sps[1].set_xlabel('Strain')
        sps[1].set_ylabel('Stress [kPa]')
        sps[0].legend()
        sps[1].legend()

        plt.show()


if __name__ == '__main__':
    run_example()



import os

import numpy as np
import o3seespy as o3

ecp2o3_type_dict = {'TAU': ['stress', 'sxy'],
                    'ESIGY': ['stress', 'syy'],
                    'ESIGX': ['stress', 'sxx'],
                    'STRS': ['strain', 'gxy']}


class O3SRAOutputs(object):
    cache_path = ''
    out_dict = None
    area = 1.0
    outs = None
    results_collected = False

    def start_recorders(self, osi, outs, sn, eles, rec_dt, sn_xy=False):
        self.rec_dt = rec_dt
        self.eles = eles
        self.sn_xy = sn_xy
        if sn_xy:
            self.nodes = sn[0, :]
            f_order = 'F'
        else:
            self.nodes = sn[:, 0]
            f_order = 'C'
        self.outs = outs
        node_depths = np.array([node.y for node in sn[:, 0]])
        ele_depths = (node_depths[1:] + node_depths[:-1]) / 2
        rd = {}
        srd = {}
        for otype in outs:
            if otype in ['ACCX', 'DISPX', 'PP']:
                if isinstance(outs[otype], str) and outs[otype] == 'all':

                    if otype == 'ACCX':
                        rd['ACCX'] = o3.recorder.NodesToArrayCache(osi, nodes=self.nodes, dofs=[o3.cc.DOF2D_X], res_type='accel',
                                                                dt=rec_dt)
                    if otype == 'DISPX':
                        rd['DISPX'] = o3.recorder.NodesToArrayCache(osi, nodes=self.nodes, dofs=[o3.cc.DOF2D_X], res_type='disp',
                                                                dt=rec_dt)
                    if otype == 'PP':
                        rd['PP'] = o3.recorder.NodesToArrayCache(osi, nodes=self.nodes, dofs=[o3.cc.DOF2D_PP], res_type='vel',
                                                                dt=rec_dt)
                else:
                    rd['ACCX'] = []
                    for i in range(len(outs['ACCX'])):
                        ind = np.argmin(abs(node_depths - outs['ACCX'][i]))
                        rd['ACCX'].append(
                            o3.recorder.NodeToArrayCache(osi, node=sn[ind][0], dofs=[o3.cc.X], res_type='accel', dt=rec_dt))
            if otype in ecp2o3_type_dict:
                rname = ecp2o3_type_dict[otype][0]  # recorder name
                for ele in eles:
                    assert isinstance(ele, o3.element.SSPquad) or isinstance(ele, o3.element.SSPquadUP)

                if isinstance(outs[otype], str) and outs[otype] == 'all':
                    if rname not in srd:
                        srd[rname] = o3.recorder.ElementsToArrayCache(osi, eles=eles, arg_vals=[rname], dt=rec_dt)
                else:
                    raise ValueError('Currently not supported')
                    # rd['STRESS'] = []
                    # for i in range(len(outs['TAU'])):
                    #     ind = np.argmin(abs(ele_depths - outs['TAU'][i]))
                    #     rd['STRESS'].append(o3.recorder.ElementToArrayCache(osi, ele=eles[ind], arg_vals=['stress'], dt=rec_dt))
            # if otype == 'ESIGY':
            #     if isinstance(outs['TAU'], str) and outs['TAU'] == 'all':
            #         if 'stress' not in srd:
            #             srd['stress'] = o3.recorder.ElementsToArrayCache(osi, eles=eles, arg_vals=['stress'], dt=rec_dt)
            if otype == 'TAUX':
                if isinstance(outs['TAUX'], str) and outs['TAUX'] == 'all':
                    rd['TAUX'] = o3.recorder.NodesToArrayCache(osi, nodes=sn.flatten(f_order), dofs=[o3.cc.X],
                                                               res_type='reaction',
                                                               dt=rec_dt)
            if otype == 'STRSX':
                if isinstance(outs['STRSX'], str) and outs['STRSX'] == 'all':
                    if 'DISPX' in outs:
                        continue
                    if sn_xy:
                        nodes = sn[0, :]
                    else:
                        nodes = sn[:, 0]
                    rd['DISPX'] = o3.recorder.NodesToArrayCache(osi, nodes=nodes, dofs=[o3.cc.X], res_type='disp',
                                                                dt=rec_dt)
        rd['TIME'] = o3.recorder.TimeToArrayCache(osi, dt=rec_dt)
        self.rd = rd
        self.srd = srd

    def results_to_files(self):
        if not self.results_collected:
            od = self.results_to_dict()
        else:
            od = self.out_dict
        for item in od:
            ffp = self.cache_path + f'{item}.txt'
            if os.path.exists(ffp):
                os.remove(ffp)
            np.savetxt(ffp, od[item])

    def load_results_from_files(self, outs=None):
        if outs is None:
            outs = ['ACCX', 'TAU', 'STRS', 'time']
        else:
            outs = list(outs)
            if 'time' not in outs:
                outs.append('time')
        od = {}
        for item in outs:
            od[item] = np.loadtxt(self.cache_path + f'{item}.txt')
        return od

    def results_to_dict(self):
        self.results_collected = True
        ro = o3.recorder.load_recorder_options()
        import pandas as pd
        df = pd.read_csv(ro)
        if self.outs is None:
            raise ValueError('outs is None')
            # items = list(self.rd)
        else:
            items = list(self.outs)

        if self.out_dict is None:
            self.out_dict = {}
            for item in self.srd:
                self.srd[item] = self.srd[item].collect().T
            for otype in items:
                if otype in self.rd:
                    vals = self.rd[otype].collect().T
                    self.out_dict[otype] = []
                    if otype == 'TAUX':
                        f_static = -np.cumsum(vals[::2, :] - vals[1::2, :], axis=0)[:-1]  # add left and right
                        f_dyn = vals[::2, :] + vals[1::2, :]  # add left and right
                        f_dyn_av = (f_dyn[1:] + f_dyn[:-1]) / 2
                        # self.out_dict[otype] = (f[1:, :] - f[:-1, :]) / area
                        self.out_dict[otype] = (f_dyn_av + f_static) / self.area
                    else:
                        self.out_dict[otype] = vals
                else:
                    if otype in ecp2o3_type_dict:
                        rname = ecp2o3_type_dict[otype][0]
                        ostr = ecp2o3_type_dict[otype][1]
                        dfe = df[df['recorder'] == rname]
                        vals = self.srd[rname]
                        cur_ind = 0
                        self.out_dict[otype] = []
                        for ele in self.eles:
                            mat_type = ele.mat.type
                            form = 'PlaneStrain'
                            dfm = dfe[(dfe['mat'] == mat_type) & (dfe['form'] == form)]
                            assert len(dfm) == 1, len(dfm)
                            outs = dfm['outs'].iloc[0].split('-')
                            oind = outs.index(ostr)
                            self.out_dict[otype].append(vals[cur_ind + oind])
                            cur_ind += len(outs)
                        self.out_dict[otype] = np.array(self.out_dict[otype])
                    if otype == 'STRSX':
                        depths = []
                        for node in self.nodes:
                            depths.append(node.y)
                        depths = np.array(depths)
                        d_incs = depths[1:] - depths[:-1]
                        vals = self.rd['DISPX'].collect(unlink=False).T
                        self.out_dict[otype] = (vals[1:] - vals[:-1]) / d_incs[:, np.newaxis]

            # Create time output
            if 'ACCX' in self.out_dict:
                self.out_dict['time'] = np.arange(0, len(self.out_dict['ACCX'][0])) * self.rec_dt
            elif 'TAU' in self.out_dict:
                self.out_dict['time'] = np.arange(0, len(self.out_dict['TAU'][0])) * self.rec_dt
        self.out_dict['TIME'] = self.rd['TIME'].collect()
        return self.out_dict


import o3seespy as o3
import numpy as np
import math


def run_ts_custom_strain(mat, esig_v0, strains, osi=None, nu_dyn=None, target_d_inc=0.00001, handle='silent', verbose=0, opyfile=None):
    k0 = 1.0
    pois = k0 / (1 + k0)
    damp = 0.05
    omega0 = 0.2
    omega1 = 20.0
    a1 = 2. * damp / (omega0 + omega1)
    a0 = a1 * omega0 * omega1
    if osi is None:
        osi = o3.OpenSeesInstance(ndm=2, ndf=2)
        mat.build(osi)

    # Establish nodes
    h_ele = 1.
    nodes = [
        o3.node.Node(osi, 0.0, 0.0, h_ele), o3.node.Node(osi, h_ele, 0.0, h_ele),  # left-bot-front -> right-bot-front
        o3.node.Node(osi, h_ele, 0.0, 0.0), o3.node.Node(osi, 0.0, 0.0, 0.0),  # right-bot-back -> left-bot-back
        o3.node.Node(osi, 0.0, h_ele, h_ele), o3.node.Node(osi, h_ele, h_ele, h_ele),  # left-top-ft -> right-top-front
        o3.node.Node(osi, h_ele, h_ele, 0.0), o3.node.Node(osi, 0.0, h_ele, 0.0)  # right-top-back -> left-top-back
    ]

    # Fix bottom nodes
    o3.Fix3DOFMulti(osi, nodes[:4], o3.cc.FIXED, o3.cc.FIXED, o3.cc.FIXED)

    # Set out-of-plane DOFs to be slaved
    o3.EqualDOFMulti(osi, nodes[4], nodes[5:], [o3.cc.X, o3.cc.Y, o3.cc.DOF3D_Z])

    ele = o3.element.SSPbrick(osi, nodes, mat, 0.0, 0.0, 0.0)

    o3.constraints.Transformation(osi)
    o3.test_check.NormDispIncr(osi, tol=1.0e-6, max_iter=35, p_flag=0)
    o3.algorithm.Newton(osi)
    o3.numberer.RCM(osi)
    o3.system.FullGeneral(osi)
    o3.integrator.DisplacementControl(osi, nodes[4], o3.cc.DOF2D_Y, 0.005)
    # o3.rayleigh.Rayleigh(osi, a0, a1, 0.0, 0.0)
    o3.analysis.Static(osi)
    o3.update_material_stage(osi, mat, stage=0)

    # Add static vertical pressure and stress bias
    ts0 = o3.time_series.Path(osi, time=[0, 1000, 1e10],
                              values=[0, 1, 1], factor=1)
    pat0 = o3.pattern.Plain(osi, ts0)
    o3.Load(osi, nodes[4], [0, -esig_v0 * h_ele / 4, 0])
    o3.Load(osi, nodes[5], [0, -esig_v0 * h_ele / 4, 0])
    o3.Load(osi, nodes[6], [0, -esig_v0 * h_ele / 4, 0])
    o3.Load(osi, nodes[7], [0, -esig_v0 * h_ele / 4, 0])
    o3.record(osi)

    o3.analyze(osi, num_inc=1000)
    stresses = o3.get_ele_response(osi, ele, 'stress')
    print('init_stress0: ', stresses)

    o3.wipe_analysis(osi)
    o3.constraints.Transformation(osi)
    o3.test_check.NormDispIncr(osi, tol=1.0e-6, max_iter=35, p_flag=0)
    o3.algorithm.Newton(osi)
    o3.numberer.RCM(osi)
    o3.system.FullGeneral(osi)
    o3.analysis.Static(osi)

    o3.update_material_stage(osi, mat, stage=1)
    o3.analyze(osi, 25, dt=1)
    if nu_dyn is not None:
        mat.set_nu(osi, nu_dyn, eles=[ele])

    # o3.extensions.to_py_file(osi)
    stresses = o3.get_ele_response(osi, ele, 'stress')
    print('init_stress1: ', stresses)

    exit_code = None
    strain = [0]
    stresses = o3.get_ele_response(osi, ele, 'stress')
    force0 = o3.get_node_reaction(osi, nodes[0], o3.cc.DOF2D_X)
    force1 = o3.get_node_reaction(osi, nodes[1], o3.cc.DOF2D_X)
    stress = [-force0 - force1]
    v_eff = [stresses[1]]
    h_eff = [stresses[0]]
    d_incs = np.diff(strains, prepend=0)
    for i in range(len(strains)):
        d_inc_i = d_incs[i]
        if target_d_inc < abs(d_inc_i):
            n = int(abs(d_inc_i / target_d_inc))
            d_step = d_inc_i / n
        else:
            n = 1
            d_step = d_inc_i
        for j in range(n):
            o3.integrator.DisplacementControl(osi, nodes[4], o3.cc.DOF2D_X, -d_step)
            o3.Load(osi, nodes[2], [1.0, 0.0, 0.0])
            o3.Load(osi, nodes[3], [1.0, 0.0, 0.0])
            o3.analyze(osi, 1)
            o3.gen_reactions(osi)
            stresses = o3.get_ele_response(osi, ele, 'stress')
            if opyfile:
                import o3seespy.extensions
                o3.extensions.to_py_file(osi, opyfile)
                opyfile = None
            print(stresses, d_step)
            v_eff.append(stresses[1])
            h_eff.append(stresses[0])
            stress.append(stresses[2])
            end_strain = o3.get_node_disp(osi, nodes[4], dof=o3.cc.DOF2D_X)
            strain.append(end_strain)

    return -np.array(stress), -np.array(strain), np.array(v_eff), np.array(h_eff), exit_code


def run_ud_custom_strain(mat, esig_v0, disps, osi=None, nu_dyn=None, target_d_inc=0.00001, handle='silent', verbose=0, opyfile=None):

    damp = 0.05
    omega0 = 0.2
    omega1 = 20.0
    a1 = 2. * damp / (omega0 + omega1)
    a0 = a1 * omega0 * omega1

    if osi is None:
        osi = o3.OpenSeesInstance(ndm=2, ndf=3)
        mat.build(osi)

    # Establish nodes
    h_ele = 1.
    nodes = [
        o3.node.Node(osi, 0.0, 0.0),
        o3.node.Node(osi, h_ele, 0.0),
        o3.node.Node(osi, h_ele, h_ele),
        o3.node.Node(osi, 0.0, h_ele)
    ]

    # Fix bottom node
    o3.Fix3DOF(osi, nodes[0], o3.cc.FIXED, o3.cc.FIXED, o3.cc.FIXED)
    o3.Fix3DOF(osi, nodes[1], o3.cc.FIXED, o3.cc.FIXED, o3.cc.FIXED)
    o3.Fix3DOF(osi, nodes[2], o3.cc.FREE, o3.cc.FREE, o3.cc.FIXED)
    o3.Fix3DOF(osi, nodes[3], o3.cc.FREE, o3.cc.FREE, o3.cc.FIXED)
    # Set out-of-plane DOFs to be slaved
    o3.EqualDOF(osi, nodes[2], nodes[3], [o3.cc.X, o3.cc.Y])

    water_bulk_mod = 2.2e6
    ele = o3.element.SSPquadUP(osi, nodes, mat, 1.0, water_bulk_mod, 1.,
                               1.0e-4, 1.0e-4, 0.6, alpha=1.0e-5, b1=0.0, b2=0.0)

    o3.constraints.Transformation(osi)
    o3.test_check.NormDispIncr(osi, tol=1.0e-6, max_iter=35, p_flag=0)
    o3.algorithm.Newton(osi)
    o3.numberer.RCM(osi)
    o3.system.FullGeneral(osi)
    o3.integrator.Newmark(osi, gamma=5. / 6, beta=4. / 9)
    o3.rayleigh.Rayleigh(osi, a0, a1, 0.0, 0.0)
    o3.analysis.Transient(osi)

    # Add static vertical pressure and stress bias
    time_series = o3.time_series.Path(osi, time=[0, 100, 1e10], values=[0, 1, 1])
    o3.pattern.Plain(osi, time_series)
    o3.Load(osi, nodes[2], [0, -esig_v0 / 2, 0])
    o3.Load(osi, nodes[3], [0, -esig_v0 / 2, 0])

    o3.analyze(osi, num_inc=110, dt=1)
    init_stresses = o3.get_ele_response(osi, ele, 'stress')
    print('init_stress0: ', init_stresses)

    ts2 = o3.time_series.Path(osi, time=[110, 80000, 1e10], values=[1., 1., 1.], factor=1)
    o3.pattern.Plain(osi, ts2, fact=1.)
    y_vert = o3.get_node_disp(osi, nodes[3], o3.cc.Y)
    o3.SP(osi, nodes[2], dof=o3.cc.Y, dof_values=[y_vert])
    o3.SP(osi, nodes[3], dof=o3.cc.Y, dof_values=[y_vert])

    # Close the drainage valves
    for node in nodes:
        o3.remove_sp(osi, node, dof=3)
    o3.analyze(osi, 25, dt=1)
    print('here3: ', o3.get_ele_response(osi, ele, 'stress'), esig_v0)

    if hasattr(mat, 'update_to_nonlinear'):
        mat.update_to_nonlinear()
        o3.analyze(osi, 25, dt=1)
    if hasattr(mat, 'set_first_call'):
        mat.set_first_call(value=0, ele=ele)
    if nu_dyn is not None:
        mat.set_nu(nu_dyn, ele=ele)
        o3.analyze(osi, 25, dt=1)

    o3.extensions.to_py_file(osi)
    curr_stresses = o3.get_ele_response(osi, ele, 'stress')
    print('init_stress1: ', curr_stresses)

    exit_code = None
    ihd = o3.get_node_disp(osi, nodes[2], dof=o3.cc.DOF2D_X)  # initial horizontal displacement
    strain = [0]
    stresses = o3.get_ele_response(osi, ele, 'stress')
    force0 = o3.get_node_reaction(osi, nodes[0], o3.cc.DOF2D_X)
    force1 = o3.get_node_reaction(osi, nodes[1], o3.cc.DOF2D_X)
    stress = [-force0 - force1]
    v_eff = [stresses[1]]
    h_eff = [stresses[0]]
    target_disps = np.array(disps) + ihd

    for i in range(len(target_disps)):
        h_disp = o3.get_node_disp(osi, nodes[2], o3.cc.X)
        curr_time = o3.get_time(osi)
        steps = int(abs(target_disps[i] - h_disp) / target_d_inc)
        ts0 = o3.time_series.Path(osi, time=[curr_time, curr_time + steps, 1e10],
                                  values=[h_disp, target_disps[i], target_disps[i]],
                                  factor=1)
        pat0 = o3.pattern.Plain(osi, ts0)
        o3.SP(osi, nodes[2], dof=o3.cc.X, dof_values=[1.0])
        for j in range(steps):
            o3.analyze(osi, 1, dt=1)
            curr_stresses = o3.get_ele_response(osi, ele, 'stress')
            if opyfile:
                import o3seespy.extensions
                o3.extensions.to_py_file(osi, opyfile)
                opyfile = None
            v_eff.append(curr_stresses[1])
            h_eff.append(curr_stresses[0])
            stress.append(curr_stresses[2])
            end_strain = o3.get_node_disp(osi, nodes[2], dof=o3.cc.DOF2D_X)
            strain.append(end_strain - ihd)

        o3.remove_load_pattern(osi, pat0)
        o3.remove(osi, ts0)
        o3.remove_sp(osi, nodes[2], dof=o3.cc.X)

    return np.array(stress), np.array(strain), -np.array(v_eff), -np.array(h_eff), exit_code


def run_ud_custom_stress(mat, esig_v0, stresses, osi=None, nu_dyn=None, target_d_inc=0.00001, handle='silent', verbose=0, opyfile=None):

    damp = 0.05
    omega0 = 0.2
    omega1 = 20.0
    a1 = 2. * damp / (omega0 + omega1)
    a0 = a1 * omega0 * omega1

    if osi is None:
        osi = o3.OpenSeesInstance(ndm=2, ndf=3)
        mat.build(osi)

    # Establish nodes
    h_ele = 1.
    nodes = [
        o3.node.Node(osi, 0.0, 0.0),
        o3.node.Node(osi, h_ele, 0.0),
        o3.node.Node(osi, h_ele, h_ele),
        o3.node.Node(osi, 0.0, h_ele)
    ]

    # Fix bottom node
    o3.Fix3DOF(osi, nodes[0], o3.cc.FIXED, o3.cc.FIXED, o3.cc.FIXED)
    o3.Fix3DOF(osi, nodes[1], o3.cc.FIXED, o3.cc.FIXED, o3.cc.FIXED)
    o3.Fix3DOF(osi, nodes[2], o3.cc.FREE, o3.cc.FREE, o3.cc.FIXED)
    o3.Fix3DOF(osi, nodes[3], o3.cc.FREE, o3.cc.FREE, o3.cc.FIXED)
    # Set out-of-plane DOFs to be slaved
    o3.EqualDOF(osi, nodes[2], nodes[3], [o3.cc.X, o3.cc.Y])

    # ele = o3.element.SSPquad(osi, nodes, mat, 'PlaneStrain', 1, 0.0, 0.0)
    water_bulk_mod = 2.2e6
    ele = o3.element.SSPquadUP(osi, nodes, mat, 1.0, water_bulk_mod, 1.,
                               1.0e-4, 1.0e-4, 0.6, alpha=1.0e-5, b1=0.0, b2=0.0)

    o3.constraints.Transformation(osi)
    o3.test_check.NormDispIncr(osi, tol=1.0e-6, max_iter=35, p_flag=0)
    o3.algorithm.Newton(osi)
    o3.numberer.RCM(osi)
    o3.system.FullGeneral(osi)
    o3.integrator.Newmark(osi, gamma=5. / 6, beta=4. / 9)
    o3.rayleigh.Rayleigh(osi, a0, a1, 0.0, 0.0)
    o3.analysis.Transient(osi)

    # Add static vertical pressure and stress bias
    # time_series = o3.time_series.Path(osi, time=[0, 100, 1e10], values=[0, 1, 1])
    # o3.pattern.Plain(osi, time_series)
    time_series = o3.time_series.Path(osi, time=[0, 100, 1e10], values=[0, 1, 1])
    o3.pattern.Plain(osi, time_series)
    o3.Load(osi, nodes[2], [0, -esig_v0 / 2, 0])
    o3.Load(osi, nodes[3], [0, -esig_v0 / 2, 0])

    o3.analyze(osi, num_inc=110, dt=1)
    init_stresses = o3.get_ele_response(osi, ele, 'stress')
    print('init_stress0: ', init_stresses)

    ts2 = o3.time_series.Path(osi, time=[110, 80000, 1e10], values=[1., 1., 1.], factor=1)
    o3.pattern.Plain(osi, ts2, fact=1.)
    y_vert = o3.get_node_disp(osi, nodes[3], o3.cc.Y)
    o3.SP(osi, nodes[2], dof=o3.cc.Y, dof_values=[y_vert])
    o3.SP(osi, nodes[3], dof=o3.cc.Y, dof_values=[y_vert])

    # Close the drainage valves
    for node in nodes:
        o3.remove_sp(osi, node, dof=3)
    o3.analyze(osi, 25, dt=1)
    print('here3: ', o3.get_ele_response(osi, ele, 'stress'), esig_v0)

    if hasattr(mat, 'update_to_nonlinear'):
        mat.update_to_nonlinear()
        o3.analyze(osi, 25, dt=1)
    if hasattr(mat, 'set_first_call'):
        mat.set_first_call(value=0, ele=ele)
    if nu_dyn is not None:
        mat.set_nu(nu_dyn, ele=ele)
        o3.analyze(osi, 25, dt=1)

    o3.extensions.to_py_file(osi)
    curr_stresses = o3.get_ele_response(osi, ele, 'stress')
    print('init_stress1: ', curr_stresses)

    exit_code = None
    strain = [0]
    curr_stresses = o3.get_ele_response(osi, ele, 'stress')
    force0 = o3.get_node_reaction(osi, nodes[0], o3.cc.DOF2D_X)
    force1 = o3.get_node_reaction(osi, nodes[1], o3.cc.DOF2D_X)
    stress = [-force0 - force1]
    v_eff = [curr_stresses[1]]
    h_eff = [curr_stresses[0]]
    prev_stress = stress[0]
    o3.record(osi)
    for i in range(len(stresses)):
        if stresses[i] >= prev_stress:
            target_disp = 0.1
        else:
            target_disp = -0.1
        if stresses[i] >= 0:
            sgn_f = 1
        else:
            sgn_f = -1

        h_disp = o3.get_node_disp(osi, nodes[2], o3.cc.X)
        curr_time = o3.get_time(osi)
        steps = 1000
        ts0 = o3.time_series.Path(osi, time=[curr_time, curr_time + steps, 1e10], values=[h_disp, target_disp, target_disp],
                                  factor=1)
        pat0 = o3.pattern.Plain(osi, ts0)
        o3.SP(osi, nodes[2], dof=o3.cc.X, dof_values=[1.0])
        curr_stress = o3.get_ele_response(osi, ele, 'stress')[2]
        if math.isnan(curr_stress):
            raise ValueError

        if opyfile:
            o3.extensions.to_py_file(osi, opyfile)
            opyfile = None
        n = 0
        while curr_stress * sgn_f < stresses[i] * sgn_f and n < 1000:  # TODO: change the strain controlled version to be like this
            n += 1
            o3.analyze(osi, 1, dt=1)
            o3.gen_reactions(osi)
            curr_stresses = o3.get_ele_response(osi, ele, 'stress')
            if opyfile:
                import o3seespy.extensions
                o3.extensions.to_py_file(osi, opyfile)
                opyfile = None
            v_eff.append(curr_stresses[1])
            h_eff.append(curr_stresses[0])
            force0 = o3.get_node_reaction(osi, nodes[0], o3.cc.DOF2D_X)
            force1 = o3.get_node_reaction(osi, nodes[1], o3.cc.DOF2D_X)
            stress.append(-force0 - force1)
            curr_stress = stress[-1]
            end_strain = o3.get_node_disp(osi, nodes[2], dof=o3.cc.DOF2D_X)
            strain.append(end_strain)
            print(curr_stress, stresses[i])
        o3.remove_load_pattern(osi, pat0)
        o3.remove(osi, ts0)
        o3.remove_sp(osi, nodes[2], dof=o3.cc.X)

    return -np.array(stress), -np.array(strain), -np.array(v_eff), -np.array(h_eff), exit_code


def example_of_run_ts_custom_strain(show=0):
    osi = o3.OpenSeesInstance(ndm=3, ndf=3, state=3)

    esig_v0 = 50.0e3
    poissons_ratio = 0.3
    g_mod = 1.0e6
    b_mod = 2 * g_mod * (1 + poissons_ratio) / (3 * (1 - 2 * poissons_ratio))

    mat = o3.nd_material.PressureIndependMultiYield(osi, 3, 2058.49, g_mod, b_mod, 68000.0, 0.1, 0.0, 100000.0, 0.0, 25)
    # mat = o3.nd_material.ElasticIsotropic(osi, e_mod=1.0e6, nu=0.3)
    ss, es = run_vload(mat, v_pressure=esig_v0, osi=osi)
    if show:
        import matplotlib.pyplot as plt
        plt.plot(es[:, 1], ss[:, 1])
        plt.show()



def example_of_run_ud_custom_strain(show=0):
    import o3seespy as o3
    import liquepy as lq

    esig_v0 = 69.131
    gravity = 9.8

    sl = lq.num.o3.PM4Sand(liq_mass_density=1.0)
    sl.relative_density = 0.35
    sl.g0_mod = 476.0
    sl.h_po = 0.53
    sl.unit_sat_weight = 1.42 * gravity

    sl.e_min = 0.5
    sl.e_max = 0.8
    sl.poissons_ratio = 0.3
    sl.phi = 33.

    sl.permeability = 1.0e-9
    sl.p_atm = 101.0

    peak_strains = [0.0001, -0.001, 0.0002]

    nu_init = sl.poissons_ratio
    pm4sand = o3.nd_material.PM4Sand(None, sl.relative_density, sl.g0_mod, sl.h_po, sl.unit_sat_mass,
                                     p_atm=101.3, nu=nu_init)

    stress, strain, vstress, hstress, ecode = run_ud_custom_strain(pm4sand, esig_v0, peak_strains)
    if show:
        import matplotlib.pyplot as plt
        bf, sps = plt.subplots(nrows=3)
        sps[0].plot(stress, label='shear')
        sps[0].plot(vstress[0] - vstress, label='PPT')
        sps[1].plot(strain, stress, label='o3seespy')
        sps[2].plot(strain, label='o3seespy')
        for ps in peak_strains:
            sps[2].axhline(ps, c='k')

        sps[0].set_xlabel('Time [s]')
        sps[0].set_ylabel('Stress [kPa]')
        sps[1].set_xlabel('Strain')
        sps[1].set_ylabel('Stress [kPa]')
        sps[0].legend()
        sps[1].legend()
        plt.show()


if __name__ == '__main__':
    example_of_run_ud_custom_strain(show=1)


import numpy as np
import o3seespy as o3
import o3seespy.extensions
import copy
import os
import o3soil
from o3soil.sra.output import O3SRAOutputs


class ESSRA1D(object):
    osi = None

    def __init__(self, sp, dy=0.5, k0=0.5, base_imp=0, cache_path=None, opfile=None, verbose=0):
        """

        Parameters
        ----------
        sp
        dy
        k0
        base_imp: float
            If positive then use as impedence at base of model,
            If zero then use last soil layer
            If negative then use fixed base
        cache_path
        opfile
        """
        self.sp = sp
        sp.gen_split(props=['shear_vel', 'unit_mass'], target=dy)
        thicknesses = sp.split["thickness"]
        self.n_node_rows = len(thicknesses) + 1
        node_depths = -np.cumsum(sp.split["thickness"])
        self.node_depths = np.insert(node_depths, 0, 0)
        self.ele_depths = (self.node_depths[1:] + self.node_depths[:-1]) / 2
        self.unit_masses = sp.split["unit_mass"] / 1e3

        self.grav = 9.81

        self.k0 = k0
        self.base_imp = base_imp

        self.ele_width = 3 * min(thicknesses)
        self.cache_path = cache_path
        self.opfile = opfile
        # Defined in static analysis
        self.soil_mats = None
        self.eles = None
        self.sn = None  # soil nodes
        self.verbose = verbose

    def build_model(self):
        # Define nodes and set boundary conditions for simple shear deformation
        # Start at top and build down?
        if self.opfile:
            self.state = 3
        else:
            self.state = 0
        if self.osi is None:
            self.osi = o3.OpenSeesInstance(ndm=2, ndf=3, state=self.state)
        nx = 1
        sn = []
        # sn = [[o3.node.Node(osi, ele_width * j, 0) for j in range(nx + 1)]]
        for i in range(0, self.n_node_rows):
            # Establish left and right nodes
            sn.append([o3.node.Node(self.osi, self.ele_width * j, self.node_depths[i]) for j in range(nx + 1)])
            # set x and y dofs equal for left and right nodes
            if i != self.n_node_rows - 1:
                o3.EqualDOF(self.osi, sn[i][0], sn[i][-1], [o3.cc.DOF2D_X, o3.cc.DOF2D_Y, o3.cc.DOF2D_PP])
        sn = np.array(sn)

        if self.base_imp < 0:
            # Fix base nodes
            for j in range(nx + 1):
                o3.Fix3DOF(self.osi, sn[-1][j], o3.cc.FIXED, o3.cc.FIXED, o3.cc.FIXED)  # Fixed pore pressure at base
        else:
            # Fix base nodes
            for j in range(nx + 1):
                o3.Fix3DOF(self.osi, sn[-1][j], o3.cc.FREE, o3.cc.FIXED, o3.cc.FIXED)  # Fixed pore pressure at base

            # Define dashpot nodes
            self.osi.reset_model_params(2, ndf=2)
            dashpot_node_l = o3.node.Node(self.osi, 0, self.node_depths[-1])
            dashpot_node_2 = o3.node.Node(self.osi, 0, self.node_depths[-1])
            o3.Fix2DOF(self.osi, dashpot_node_l, o3.cc.FIXED, o3.cc.FIXED)
            o3.Fix2DOF(self.osi, dashpot_node_2, o3.cc.FREE, o3.cc.FIXED)

            # define equal DOF for dashpot and soil base nodes
            o3.EqualDOF(self.osi, sn[-1][0], sn[-1][1], [o3.cc.X])
            o3.EqualDOF(self.osi, sn[-1][0], dashpot_node_2, [o3.cc.X])

        # define materials
        pois = self.k0 / (1 + self.k0)
        ele_thick = 1.0  # m
        self.soil_mats = []
        strains = np.logspace(-6, -0.5, 16)
        prev_args = []
        prev_kwargs = {}
        prev_sl_class = None
        self.eles = []
        for i in range(len(self.ele_depths)):
            y_depth = -self.ele_depths[i]

            sl_id = self.sp.get_layer_index_by_depth(y_depth)
            sl = self.sp.layer(sl_id)

            if hasattr(sl, 'op_type'):
                if sl.built:
                    pass
                else:
                    sl.build(self.osi)
                    mat = sl
                    self.soil_mats.append(mat)
            else:
                if y_depth > self.sp.gwl:
                    saturated = True
                else:
                    saturated = False
                esig_v0 = self.sp.get_v_eff_stress_at_depth(y_depth)
                sl_class, args, kwargs = o3soil.get_o3_class_and_args_from_soil_obj(sl, saturated,
                                                                                    overrides={'nu': pois},
                                                                                    esig_v0=esig_v0)
                if o3.extensions.has_o3_model_changed(sl_class, prev_sl_class, args, prev_args, kwargs, prev_kwargs):
                    mat = sl_class(self.osi, *args, **kwargs)
                    prev_sl_class = sl_class
                    prev_args = copy.deepcopy(args)
                    prev_kwargs = copy.deepcopy(kwargs)
                    mat.dynamic_poissons_ratio = sl.poissons_ratio
                    self.soil_mats.append(mat)

            # def element
            if hasattr(self.sp, 'water_bulk_mod') and self.sp.water_bulk_mod is not None:
                k_water = self.sp.water_bulk_mod
            else:
                k_water = 2.2e6
            for xx in range(nx):  # TODO: account for gwl
                nodes = [sn[i + 1][xx], sn[i + 1][xx + 1], sn[i][xx + 1], sn[i][xx]]  # anti-clockwise
                a_sspquad_up = 6.0e-5
                self.eles.append(o3.element.SSPquadUP(self.osi, nodes, mat, ele_thick, k_water, f_den=1.0, k1=sl.permeability,
                                                      k2=sl.permeability, void=sl.e_curr, alpha=a_sspquad_up,
                                                      b2=-self.grav))
        self.sn = sn
        if self.base_imp >= 0:
            # define material and element for viscous dampers
            if self.base_imp == 0:
                sl = self.sp.get_soil_at_depth(self.sp.height)
                base_imp = sl.unit_dry_mass * self.sp.get_shear_vel_at_depth(self.sp.height)
            self.c_base = self.ele_width * base_imp / 1e3
            dashpot_mat = o3.uniaxial_material.Viscous(self.osi, self.c_base, alpha=1.)
            o3.element.ZeroLength(self.osi, [dashpot_node_l, dashpot_node_2], mats=[dashpot_mat], dirs=[o3.cc.DOF2D_X])

        self.o3res = o3.results.Results2D(cache_path=self.cache_path)
        self.o3res.wipe_old_files()
        self.o3res.coords = o3.get_all_node_coords(self.osi)
        self.o3res.ele2node_tags = o3.get_all_ele_node_tags_as_dict(self.osi)
        self.o3res.mat2ele_tags = []
        for ele in self.eles:
            self.o3res.mat2ele_tags.append([ele.mat.tag, ele.tag])

    def execute_static(self, ray_freqs=(0.5, 10), xi=0.03):
        # Static analysis
        # for i in range(len(self.soil_mats)):  # TODO: should be a method on object 'update_to_linear'
        #     o3.update_material_stage(self.osi, self.soil_mats[i], 0)
        o3.constraints.Transformation(self.osi)
        o3.test.NormDispIncr(self.osi, tol=1.0e-5, max_iter=30, p_flag=0)
        o3.algorithm.KrylovNewton(self.osi)
        o3.numberer.RCM(self.osi)
        o3.system.ProfileSPD(self.osi)
        o3.integrator.Newmark(self.osi, gamma=0.5, beta=0.25)
        o3.analysis.Transient(self.osi)
        omega_1 = 2 * np.pi * ray_freqs[0]
        omega_2 = 2 * np.pi * ray_freqs[1]
        a0 = 2 * xi * omega_1 * omega_2 / (omega_1 + omega_2)
        a1 = 2 * xi / (omega_1 + omega_2)
        # o3.rayleigh.Rayleigh(self.osi, a0, a1, 0, 0)
        o3.analyze(self.osi, 1000, 5.)
        # if self.opfile:
        #     o3.extensions.to_py_file(self.osi, self.opfile, compress=True)
            # o3.extensions.to_tcl_file(self.osi, self.opfile.replace('.py', '.tcl'))

        for i in range(len(self.soil_mats)):
            if hasattr(self.soil_mats[i], 'update_to_nonlinear'):  # TODO: should this pass the ele number to it?
                self.soil_mats[i].update_to_nonlinear()

        for ele in self.eles:
            mat = ele.mat
            if hasattr(mat, 'set_nu'):
                mat.set_nu(mat.dynamic_poissons_ratio, eles=[ele])
            if hasattr(mat, 'set_first_call'):
                mat.set_first_call(value=0, ele=ele)
                # TODO: set_dynamic permeability
        o3.analyze(self.osi, 40, 500.)

        # reset time and analysis
        o3.wipe_analysis(self.osi)
        self.o3res.coords = o3.get_all_node_coords(self.osi)
        # if self.opfile:
        #     o3.extensions.to_py_file(self.osi, self.opfile)

    def get_nearest_node_layer_at_depth(self, depth):
        # Convert to positive since node depths go downwards
        return int(np.round(np.interp(depth, -self.node_depths, np.arange(len(self.node_depths)))))

    def get_nearest_ele_layer_at_depth(self, depth):
        # Convert to positive since ele depths go downwards
        return int(np.round(np.interp(depth, -self.ele_depths, np.arange(len(self.ele_depths)))))

    def apply_loads(self, ray_freqs=(0.5, 10), xi=0.03):
        o3.set_time(self.osi, 0.0)

        # Define the dynamic analysis
        o3.constraints.Transformation(self.osi)
        o3.test.NormDispIncr(self.osi, tol=1.0e-4, max_iter=30, p_flag=0)
        # o3.test_check.EnergyIncr(self.osi, tol=1.0e-6, max_iter=30)
        o3.algorithm.Newton(self.osi)
        o3.system.SparseGeneral(self.osi)
        o3.numberer.RCM(self.osi)
        o3.integrator.Newmark(self.osi, gamma=0.5, beta=0.25)
        o3.analysis.Transient(self.osi)
        omega_1 = 2 * np.pi * ray_freqs[0]
        omega_2 = 2 * np.pi * ray_freqs[1]
        a0 = 2 * xi * omega_1 * omega_2 / (omega_1 + omega_2)
        a1 = 2 * xi / (omega_1 + omega_2)
        o3.rayleigh.Rayleigh(self.osi, a0, a1, 0, 0)

        static_time = 500
        print('time: ', o3.get_time(self.osi))
        # Add static stress bias
        time_series = o3.time_series.Path(self.osi, time=[0, static_time / 2, static_time, 1e3], values=[0, 0.5, 1, 1],
                                          use_last=True)
        o3.pattern.Plain(self.osi, time_series)
        net_hload = 0
        for i in range(len(self.sp.hloads)):
            pload = self.sp.hloads[i].p_x
            y = self.sp.hloads[i].y
            ind = self.get_nearest_node_layer_at_depth(y)
            print(i, y, ind)
            if self.sp.loads_are_stresses:
                pload *= self.ele_width
            o3.Load(self.osi, self.sn[ind][0], [pload, 0, 0])
            net_hload += pload
        if self.base_imp >= 0:
            o3.Load(self.osi, self.sn[-1][0], [-net_hload, 0, 0])

        static_dt = 0.1
        o3.analyze(self.osi, int(static_time / static_dt), static_dt)
        o3.load_constant(self.osi, time=0)

    def execute_dynamic(self, asig, analysis_dt=0.001, ray_freqs=(0.5, 10), xi=0.03, analysis_time=None,
                        outs=None, rec_dt=None, playback_dt=None, playback=True):
        self.rec_dt = rec_dt
        self.playback_dt = playback_dt
        if rec_dt is None:
            self.rec_dt = asig.dt
        if playback_dt is None:
            self.playback_dt = analysis_dt
        o3.set_time(self.osi, 0.0)

        # Define the dynamic analysis
        o3.constraints.Transformation(self.osi)
        o3.test.NormDispIncr(self.osi, tol=1.0e-4, max_iter=30, p_flag=0)
        # o3.test_check.EnergyIncr(self.osi, tol=1.0e-6, max_iter=30)
        o3.algorithm.Newton(self.osi)
        o3.system.SparseGeneral(self.osi)
        o3.numberer.RCM(self.osi)
        o3.integrator.Newmark(self.osi, gamma=0.5, beta=0.25)
        o3.analysis.Transient(self.osi)
        # Rayleigh damping parameters
        omega_1 = 2 * np.pi * ray_freqs[0]
        omega_2 = 2 * np.pi * ray_freqs[1]
        a0 = 2 * xi * omega_1 * omega_2 / (omega_1 + omega_2)
        a1 = 2 * xi / (omega_1 + omega_2)
        o3.rayleigh.Rayleigh(self.osi, a0, a1, 0, 0)

        init_time = o3.get_time(self.osi)
        if playback:
            self.o3res.dynamic = True
            self.o3res.start_recorders(self.osi, dt=self.playback_dt)
        else:
            self.o3res.dynamic = False
        self.o3sra_outs = O3SRAOutputs()
        self.o3sra_outs.start_recorders(self.osi, outs, self.sn, self.eles, rec_dt=self.rec_dt)

        # Define the dynamic input motion
        if self.base_imp < 0:  # fixed base
            acc_series = o3.time_series.Path(self.osi, dt=asig.dt, values=asig.values)
            o3.pattern.UniformExcitation(self.osi, dir=o3.cc.X, accel_series=acc_series)
        else:
            ts_obj = o3.time_series.Path(self.osi, dt=asig.dt, values=asig.velocity * 1, factor=self.c_base)
            o3.pattern.Plain(self.osi, ts_obj)
            o3.Load(self.osi, self.sn[-1][0], [1., 0., 0])
        if self.state == 3:
            o3.extensions.to_py_file(self.osi, self.opfile, compress=True)
        # Run the dynamic motion
        o3.record(self.osi)
        curr_time = o3.get_time(self.osi)
        while curr_time - init_time < analysis_time:
            if self.verbose:
                print('time: ', curr_time)
            if o3.analyze(self.osi, 1, analysis_dt):
                print('failed')
                if o3.analyze(self.osi, 10, analysis_dt / 10):
                    break
            curr_time = o3.get_time(self.osi)
        o3.wipe(self.osi)
        self.out_dict = self.o3sra_outs.results_to_dict()

        if self.cache_path:
            self.o3sra_outs.cache_path = self.cache_path
            self.o3sra_outs.results_to_files()
            self.o3res.save_to_cache()


def run_essra(sp, asig, ray_freqs=(0.5, 10), xi=0.03, analysis_dt=0.001, dy=0.5, analysis_time=None, outs=None,
                  base_imp=0, k0=0.5, cache_path=None, opfile=None, playback=False, rec_dt=None, verbose=0):
    """

    Parameters
    ----------
    sp
    asig
    ray_freqs
    xi
    analysis_dt
    dy
    analysis_time
    outs
    base_imp: float
        If positive then use as impedence at base of model,
        If zero then use last soil layer
        If negative then use fixed base
    k0
    cache_path
    opfile
    playback

    Returns
    -------

    """
    sra_1d = ESSRA1D(sp, dy=dy, k0=k0, base_imp=base_imp, cache_path=cache_path, opfile=opfile, verbose=verbose)
    sra_1d.build_model()
    sra_1d.execute_static()
    if hasattr(sra_1d.sp, 'hloads'):
        sra_1d.apply_loads()
    sra_1d.execute_dynamic(asig, analysis_dt=analysis_dt, ray_freqs=ray_freqs, xi=xi, analysis_time=analysis_time,
                           outs=outs, playback=playback, playback_dt=0.01, rec_dt=rec_dt)
    return sra_1d




import geofound as gf
import o3seespy as o3
import numpy as np


def set_bnwf2d_via_harden_2009(osi, sl, fd, soil_node, bd_node, ip_axis, axis=None, dettach=True, soil_nl=True):
    """
    Set a Beam on nonlinear Winker Foundation between two nodes

    Parameters
    ----------
    osi
    sl: sfsimodels.Soil object
    fd: sfsimodels.Foundation object
    soil_node: o3.Node
        The soil node
    bd_node: o3.Node
        The base of the building node
    ip_axis: str
        The axis that is in the plane of deformation
    axis: str
        The axis which the foundation would rotate around (deprecated in favour of ip_axis)
    dettach
    soil_nl: bool

    Returns
    -------

    """
    if axis is not None:
        if axis == 'length':
            ip_axis = 'width'
        else:
            ip_axis = 'length'
    # TODO: account for foundation height
    end_zone_ratio = 0.3  # TODO: currently based on 0.3, but Harden et al. (2005) showed this to be a ratio of B/L
    k_rot = gf.stiffness.calc_rotational_via_gazetas_1991(sl, fd, ip_axis=ip_axis)
    k_vert = gf.stiffness.calc_vert_via_gazetas_1991(sl, fd)
    if ip_axis == 'width':  # rotation around the length axis
        k_vert_i = k_vert / fd.width / fd.length  # TODO: should the be only divide by width?
    else:
        k_vert_i = k_vert / fd.length / fd.length
    if fd.i_ww >= fd.i_ll:
        len_dominant = True
        l = fd.length * 0.5
        b = fd.width * 0.5
    else:
        len_dominant = False
        l = fd.width * 0.5
        b = fd.length * 0.5
    if (ip_axis == 'width' and len_dominant) or (ip_axis == 'length' and not len_dominant):
        # rotation around x-axis
        r_k = (3 * k_rot / (4 * k_vert_i * b ** 3 * l)) - (1 - end_zone_ratio) ** 3 / (1 - (1 - end_zone_ratio) ** 3)
    else:
        # rotation around y-axis
        r_k = (3 * k_rot / (4 * k_vert_i * b * l ** 3)) - (1 - end_zone_ratio) ** 3 / (1 - (1 - end_zone_ratio) ** 3)

    n_springs = 10
    k_spring = k_vert / n_springs
    # use 10 springs - have the exterior spring
    w = fd.width
    pos = np.linspace(w / 20, fd.width - w / 20, n_springs) - fd.width / 2
    if not soil_nl:
        if dettach:
            k_ten = 1.0e-5 * k_spring
        else:
            k_ten = k_spring
        int_spring_mat = o3.uniaxial_material.Elastic(osi, k_spring, eneg=k_ten)
        ext_spring_mat = o3.uniaxial_material.Elastic(osi, r_k * k_spring, eneg=r_k * k_ten)
    else:
        q_ult = gf.capacity_salgado_2008(sl, fd)
        f_ult = q_ult * fd.area
        f_spring = f_ult / n_springs / 1.3  # TODO: should exterior be different? -should be divide by n_springs, end should be different
        int_spring_mat_1 = o3.uniaxial_material.SteelMPF(osi, f_spring, f_spring, k_spring, 0.05, 0.05, params=[5, 0.925, 0.15])
        ext_spring_mat_1 = o3.uniaxial_material.SteelMPF(osi, f_spring, f_spring, r_k * k_spring, 0.05, 0.05, params=[5, 0.925, 0.15])
        if dettach:
            mat_obj2 = o3.uniaxial_material.Elastic(osi, 1000 * k_spring, eneg=0.0001 * k_spring)
            int_spring_mat = o3.uniaxial_material.Series(osi, [int_spring_mat_1, mat_obj2])
            ext_spring_mat = o3.uniaxial_material.Series(osi, [ext_spring_mat_1, mat_obj2])
        else:
            int_spring_mat = int_spring_mat_1
            ext_spring_mat = ext_spring_mat_1
    spring_mats = 3 * [ext_spring_mat] + 4 * [int_spring_mat] + 3 * [ext_spring_mat]
    scaler = np.ones_like(pos)
    # scaler[:3] = r_k
    # scaler[-3:] = r_k
    # k_rot_bnwf = np.sum(k_spring * scaler * pos ** 2)
    # print(k_rot, k_rot_bnwf, k_rot_bnwf / k_rot)
    from o3seespy.command.element.soil_foundation import gen_shallow_foundation_bnwf
    fd_area = fd.width * fd.height
    fd_emod = 30.0e9
    fd_iz = fd.width * fd.height ** 3 / 12
    return gen_shallow_foundation_bnwf(osi, soil_node, bd_node, sf_mats=spring_mats, pos=pos, fd_area=fd_area,
                                fd_e_mod=fd_emod, fd_iz=fd_iz)


def generate_example_ssi_system():
    import sfsimodels as sm
    sl = sm.StressDependentSoil()
    sl.e_max = 0.877
    sl.e_min = 0.511
    sl.specific_gravity = 2.65
    sl.relative_density = 0.38
    sl.phi = 32.
    sl.cohesion = 0.0
    cf = 0.77  # Arulmoli et al. (1992)
    sl.g0_mod = 625 * cf * np.sqrt(100e3) / (0.3 + 0.7 * sl.e_curr) * np.sqrt(sl.p_atm) / sl.p_atm
    sl.a = 0.5
    sl.g_mod = 23.5e6
    print('g_below f', sl.get_g_mod_at_m_eff_stress(13.7e3))
    sl.poissons_ratio = 0.3
    sp = sm.SoilProfile()
    sp.add_layer(0, sl)
    # Add steel base
    h_base = 0.183 * 49  # note N=49g
    sp.height = h_base + 0.1
    steel = sm.Soil()
    steel.g_mod = 80e9
    steel.e_curr = 0.0
    steel.xi = 0.01
    steel.poissons_ratio = 0.28
    steel.unit_sat_weight = 76930.0
    sp.add_layer(h_base, steel)

    fd = sm.RaftFoundation()
    fd.width = 7.35
    fd.length = 4.7
    fd.height = 1.24
    fd.depth = 2.24  # m
    fd.mass = 79.0e3  # kg

    bd = sm.SDOFBuilding()
    bd.h_eff = 12.1
    bd.mass_eff = 553.0e3
    bd.xi = 0.05
    bd.t_fixed = 0.05  # s
    bd.inputs.append('xi')
    bd.set_foundation(fd, x=0)
    return bd, sl


def run_example():
    osi = o3.OpenSeesInstance(ndm=2, state=3)
    bd, sl = generate_example_ssi_system()
    fd = bd.fd
    height = bd.h_eff
    # Establish nodes
    bot_node = o3.node.Node(osi, 0, 0)
    top_node = o3.node.Node(osi, 0, height)
    sl_node = o3.node.Node(osi, 0, 0)  # TODO: add fd height

    # Fix bottom node
    o3.Fix3DOF(osi, top_node, o3.cc.FREE, o3.cc.FREE, o3.cc.FREE)
    o3.Fix3DOF(osi, bot_node, o3.cc.FREE, o3.cc.FREE, o3.cc.FREE)
    o3.Fix3DOF(osi, sl_node, o3.cc.FIXED, o3.cc.FIXED, o3.cc.FIXED)
    # Set out-of-plane DOFs to be slaved
    # o3.EqualDOF(osi, top_node, bot_node, [o3.cc.Y])

    # nodal mass (weight / g):
    o3.Mass(osi, top_node, bd.mass_eff, 0., 0.)
    o3.Mass(osi, bot_node, fd.mass, 0., 0.)  # TODO: add

    # Define material
    transf = o3.geom_transf.Linear2D(osi, [])
    area = 1.0
    e_mod = 100.0e6
    iz = bd.k_eff * height ** 3 / (3 * e_mod)
    ele_nodes = [bot_node, top_node]

    vert_ele = o3.element.ElasticBeamColumn2D(osi, ele_nodes, area=area, e_mod=e_mod, iz=iz, transf=transf)

    # TODO: if sl.g_mod is stress dependent, then account for foundation load and depth increase using pg2-18 of NIST
    # TODO: Implement the stiffness using soil_profile into geofound that accounts for fd.q_load
    k_shear = gf.stiffness.calc_shear_via_gazetas_1991(sl, fd, axis='length')
    shear_mat = o3.uniaxial_material.Elastic(osi, k_shear)
    # sl.override('g_mod', sl.g_mod)
    soil_fd_ele = o3.element.ZeroLength(osi, [sl_node, bot_node], mats=[shear_mat], dirs=[o3.cc.DOF2D_X])
    bnwf = set_bnwf_via_harden_2009(osi, sl, fd, sl_node, bot_node, axis='width', soil_nl=True, dettach=True)
    import o3seespy.extensions
    o3.extensions.to_py_file(osi)

    # set damping based on first eigen mode
    angular_freq = o3.get_eigen(osi, solver='fullGenLapack', n=1)[0] ** 0.5
    response_period = 2 * np.pi / angular_freq
    print('response_period: ', response_period)
    beta_k = 2 * bd.xi / angular_freq
    o3.rayleigh.Rayleigh(osi, alpha_m=0.0, beta_k=beta_k, beta_k_init=0.0, beta_k_comm=0.0)
    ts0 = o3.time_series.Linear(osi, factor=1)
    o3.pattern.Plain(osi, ts0)
    o3.Load(osi, top_node, [0, -bd.mass_eff * 9.8, 0])
    o3.Load(osi, bot_node, [0, -fd.mass * 9.8, 0])

    o3.constraints.Transformation(osi)
    o3.test_check.NormDispIncr(osi, tol=1.0e-6, max_iter=35, p_flag=0)
    o3.algorithm.Newton(osi)
    o3.numberer.RCM(osi)
    o3.system.FullGeneral(osi)
    n_steps_gravity = 10
    d_gravity = 1. / n_steps_gravity
    o3.integrator.LoadControl(osi, d_gravity, num_iter=10)
    # o3.rayleigh.Rayleigh(osi, a0, a1, 0.0, 0.0)
    o3.analysis.Static(osi)
    o3.analyze(osi, num_inc=n_steps_gravity)
    o3.load_constant(osi, time=0.0)
    print('init_disp: ', o3.get_node_disp(osi, bot_node, o3.cc.DOF2D_Y))

    # Start horizontal load
    ts0 = o3.time_series.Linear(osi, factor=1)
    o3.pattern.Plain(osi, ts0)
    o3.Load(osi, top_node, [bd.mass_eff * 2.5, 0, 0])
    o3.constraints.Transformation(osi)
    o3.test_check.NormDispIncr(osi, tol=1.0e-6, max_iter=35, p_flag=0)
    o3.algorithm.Newton(osi)
    o3.numberer.RCM(osi)
    o3.system.FullGeneral(osi)
    n_steps_hload = 100
    d_hload = 1. / n_steps_hload
    o3.integrator.LoadControl(osi, d_hload, num_iter=10)
    # o3.rayleigh.Rayleigh(osi, a0, a1, 0.0, 0.0)
    o3.analysis.Static(osi)
    rot = []
    mom = []
    for i in range(n_steps_hload):
        o3.analyze(osi, num_inc=1)
        rot.append(o3.get_node_disp(osi, bot_node, o3.cc.DOF2D_ROTZ))
        o3.gen_reactions(osi)
        mom.append(o3.get_ele_response(osi, vert_ele, 'force')[2])

    o3.gen_reactions(osi)
    print('node disps')
    for node in bnwf.top_nodes:
        print(o3.get_node_disp(osi, node, o3.cc.DOF2D_Y))
    print('node react')
    for node in bnwf.bot_nodes:
        print(o3.get_node_reaction(osi, node, o3.cc.DOF2D_Y))

    import matplotlib.pyplot as plt
    plt.plot(rot, mom)

    plt.show()

if __name__ == '__main__':
    run_example()


from setuptools import setup, find_packages

about = {}
with open("o3soil/__about__.py") as fp:
    exec(fp.read(), about)

with open('README.rst') as readme_file:
    readme = readme_file.read()

# with open('HISTORY.rst') as history_file:
#     history = history_file.read()

setup(name='o3soil',
      version=about['__version__'],
      description='A toolkit of extensions for using o3seespy for geotechnical simulations',
      long_description=readme,  # + '\n\n' + history,
      url='',
      author=about['__author__'],
      author_email='maxim.millen@canterbury.ac.nz',
      license=about['__license__'],
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Intended Audience :: Science/Research',
          'Topic :: Scientific/Engineering',
          'Programming Language :: Python :: 3',
      ],
      packages=find_packages(exclude=['contrib', 'docs', 'tests']),
      install_requires=[
        "numpy",
        "o3seespy"
      ],
      # List additional groups of dependencies here (e.g. development
      # dependencies). You can install these using the following syntax,
      # for example:
      # $ pip install -e .[dev,test]
      extras_require={
          'test': ['pytest'],
      },
      python_requires='>=3',
      package_data={},
      zip_safe=False)


# From python packaging guides
# versioning is a 3-part MAJOR.MINOR.MAINTENANCE numbering scheme,
# where the project author increments:

# MAJOR version when they make incompatible API changes,
# MINOR version when they add functionality in a backwards-compatible manner, and
# MAINTENANCE version when they make backwards-compatible bug fixes.


import numpy as np
import sfsimodels as sm
import o3seespy as o3
import o3seespy.extensions
import copy
from o3soil.sra.output import O3SRAOutputs


class SRA1D(object):
    osi = None

    def __init__(self, sp, dy=0.5, k0=0.5, base_imp=0, cache_path=None, opfile=None):
        """

        Parameters
        ----------
        sp
        dy
        k0
        base_imp: float
            If positive then use as impedence at base of model,
            If zero then use last soil layer
            If negative then use fixed base
        cache_path
        opfile
        """
        self.sp = sp
        sp.gen_split(props=['shear_vel', 'unit_mass'], target=dy)
        thicknesses = sp.split["thickness"]
        self.n_node_rows = len(thicknesses) + 1
        node_depths = -np.cumsum(sp.split["thickness"])
        self.node_depths = np.insert(node_depths, 0, 0)
        self.ele_depths = (self.node_depths[1:] + self.node_depths[:-1]) / 2
        self.unit_masses = sp.split["unit_mass"] / 1e3

        self.grav = 9.81

        self.k0 = k0
        self.base_imp = base_imp

        self.ele_width = 3 * min(thicknesses)
        self.cache_path = cache_path
        self.opfile = opfile
        # Defined in static analysis
        self.soil_mats = None
        self.eles = None
        self.sn = None  # soil nodes

    def build_model(self):
        # Define nodes and set boundary conditions for simple shear deformation
        # Start at top and build down?
        if self.opfile:
            self.state = 3
        else:
            self.state = 0
        if self.osi is None:
            self.osi = o3.OpenSeesInstance(ndm=2, ndf=2, state=self.state)
        nx = 1
        sn = []
        # sn = [[o3.node.Node(osi, ele_width * j, 0) for j in range(nx + 1)]]
        for i in range(0, self.n_node_rows):
            # Establish left and right nodes
            sn.append([o3.node.Node(self.osi, self.ele_width * j, self.node_depths[i]) for j in range(nx + 1)])
            # set x and y dofs equal for left and right nodes
            if i != self.n_node_rows - 1:
                o3.EqualDOF(self.osi, sn[i][0], sn[i][-1], [o3.cc.DOF2D_X, o3.cc.DOF2D_Y])
        sn = np.array(sn)

        if self.base_imp < 0:
            # Fix base nodes
            for j in range(nx + 1):
                o3.Fix2DOF(self.osi, sn[-1][j], o3.cc.FIXED, o3.cc.FIXED)
        else:
            # Fix base nodes
            for j in range(nx + 1):
                o3.Fix2DOF(self.osi, sn[-1][j], o3.cc.FREE, o3.cc.FIXED)

            # Define dashpot nodes
            dashpot_node_l = o3.node.Node(self.osi, 0, self.node_depths[-1])
            dashpot_node_2 = o3.node.Node(self.osi, 0, self.node_depths[-1])
            o3.Fix2DOF(self.osi, dashpot_node_l, o3.cc.FIXED, o3.cc.FIXED)
            o3.Fix2DOF(self.osi, dashpot_node_2, o3.cc.FREE, o3.cc.FIXED)

            # define equal DOF for dashpot and soil base nodes
            o3.EqualDOF(self.osi, sn[-1][0], sn[-1][1], [o3.cc.X])
            o3.EqualDOF(self.osi, sn[-1][0], dashpot_node_2, [o3.cc.X])

        # define materials
        pois = self.k0 / (1 + self.k0)
        ele_thick = 1.0  # m
        self.soil_mats = []
        strains = np.logspace(-6, -0.5, 16)
        prev_args = []
        prev_kwargs = {}
        prev_sl_class = None
        self.eles = []
        for i in range(len(self.ele_depths)):
            y_depth = -self.ele_depths[i]

            sl_id = self.sp.get_layer_index_by_depth(y_depth)
            sl = self.sp.layer(sl_id)
            if hasattr(sl, 'op_type'):
                if sl.built:
                    pass
                else:
                    sl.build(self.osi)
                    mat = sl
                    self.soil_mats.append(mat)
            else:
                app2mod = {}
                if y_depth > self.sp.gwl:
                    umass = sl.unit_sat_mass / 1e3  # TODO: work out how to run in Pa, N, m, s
                else:
                    umass = sl.unit_dry_mass / 1e3
                overrides = {'nu': pois, 'p_atm': 101,
                             'rho': umass,
                             'unit_moist_mass': umass,
                             'nd': 2.0,
                             # 'n_surf': 25
                             }
                # Define material
                if not hasattr(sl, 'o3_type'):
                    sl.o3_type = sl.type  # for backward compatibility
                if sl.o3_type == 'pm4sand':
                    sl_class = o3.nd_material.PM4Sand
                    # overrides = {'nu': pois, 'p_atm': 101, 'unit_moist_mass': umass}
                    app2mod = sl.app2mod
                elif sl.o3_type == 'sdmodel':
                    sl_class = o3.nd_material.StressDensity
                    # overrides = {'nu': pois, 'p_atm': 101, 'unit_moist_mass': umass}
                    app2mod = sl.app2mod
                elif sl.o3_type in ['pimy', 'pdmy', 'pdmy02']:
                    if hasattr(sl, 'get_g_mod_at_m_eff_stress'):
                        if hasattr(sl, 'g_mod_p0') and sl.g_mod_p0 != 0.0:
                            v_eff = self.sp.get_v_eff_stress_at_depth(y_depth)
                            k0 = sl.poissons_ratio / (1 - sl.poissons_ratio)
                            m_eff = v_eff * (1 + 2 * k0) / 3
                            p = m_eff  # Pa
                            overrides['d'] = 0.0
                        else:
                            p = 101.0e3  # Pa
                            overrides['d'] = sl.a
                        g_mod_r = sl.get_g_mod_at_m_eff_stress(p) / 1e3
                    else:
                        p = 101.0e3  # Pa
                        overrides['d'] = 0.0
                        g_mod_r = sl.g_mod / 1e3

                    b_mod = 2 * g_mod_r * (1 + sl.poissons_ratio) / (3 * (1 - 2 * sl.poissons_ratio))
                    overrides['p_ref'] = p / 1e3
                    overrides['g_mod_ref'] = g_mod_r
                    overrides['bulk_mod_ref'] = b_mod
                    if sl.o3_type == 'pimy':
                        overrides['cohesion'] = sl.cohesion / 1e3
                        sl_class = o3.nd_material.PressureIndependMultiYield
                    elif sl.o3_type == 'pdmy':
                        sl_class = o3.nd_material.PressureDependMultiYield
                    elif sl.o3_type == 'pdmy02':
                        sl_class = o3.nd_material.PressureDependMultiYield02
                else:
                    g_mod = self.sp.split['shear_vel'][i] ** 2 / self.unit_masses[i]
                    sl_class = o3.nd_material.ElasticIsotropic
                    sl.e_mod = 2 * g_mod * (1 - sl.poissons_ratio)
                    overrides['nu'] = sl.poissons_ratio
                    app2mod['rho'] = 'unit_moist_mass'
                args, kwargs = o3.extensions.get_o3_kwargs_from_obj(sl, sl_class, custom=app2mod, overrides=overrides)

                if o3.extensions.has_o3_model_changed(sl_class, prev_sl_class, args, prev_args, kwargs, prev_kwargs):
                    mat = sl_class(self.osi, *args, **kwargs)
                    prev_sl_class = sl_class
                    prev_args = copy.deepcopy(args)
                    prev_kwargs = copy.deepcopy(kwargs)
                    mat.dynamic_poissons_ratio = sl.poissons_ratio
                    self.soil_mats.append(mat)

            # def element
            for xx in range(nx):
                nodes = [sn[i + 1][xx], sn[i + 1][xx + 1], sn[i][xx + 1], sn[i][xx]]  # anti-clockwise
                # eles.append(o3.element.Quad(self.osi, nodes, ele_thick, o3.cc.PLANE_STRAIN, mat, b2=-grav * unit_masses[i]))
                self.eles.append(o3.element.SSPquad(self.osi, nodes, mat, o3.cc.PLANE_STRAIN, ele_thick, 0.0,
                                                    -self.grav * self.unit_masses[i]))
        self.sn = sn
        if self.base_imp >= 0:
            # define material and element for viscous dampers
            if self.base_imp == 0:
                sl = self.sp.get_soil_at_depth(self.sp.height)
                base_imp = sl.unit_dry_mass * self.sp.get_shear_vel_at_depth(self.sp.height)
            self.c_base = self.ele_width * base_imp / 1e3
            dashpot_mat = o3.uniaxial_material.Viscous(self.osi, self.c_base, alpha=1.)
            o3.element.ZeroLength(self.osi, [dashpot_node_l, dashpot_node_2], mats=[dashpot_mat], dirs=[o3.cc.DOF2D_X])

        self.o3res = o3.results.Results2D(cache_path=self.cache_path)
        self.o3res.wipe_old_files()
        self.o3res.coords = o3.get_all_node_coords(self.osi)
        self.o3res.ele2node_tags = o3.get_all_ele_node_tags_as_dict(self.osi)
        self.o3res.mat2ele_tags = []
        for ele in self.eles:
            self.o3res.mat2ele_tags.append([ele.mat.tag, ele.tag])

    def execute_static(self, ray_freqs=(0.5, 10), xi=0.03):
        # Static analysis
        o3.constraints.Transformation(self.osi)
        o3.test.NormDispIncr(self.osi, tol=1.0e-5, max_iter=30, p_flag=0)
        o3.algorithm.Newton(self.osi)
        o3.numberer.RCM(self.osi)
        o3.system.ProfileSPD(self.osi)
        o3.integrator.Newmark(self.osi, gamma=0.5, beta=0.25)
        o3.analysis.Transient(self.osi)
        omega_1 = 2 * np.pi * ray_freqs[0]
        omega_2 = 2 * np.pi * ray_freqs[1]
        a0 = 2 * xi * omega_1 * omega_2 / (omega_1 + omega_2)
        a1 = 2 * xi / (omega_1 + omega_2)
        #o3.rayleigh.Rayleigh(self.osi, a0, a1, 0, 0)
        o3.analyze(self.osi, 10, 500.)
        if self.opfile:
            o3.extensions.to_py_file(self.osi, self.opfile)
            o3.extensions.to_tcl_file(self.osi, self.opfile.replace('.py', '.tcl'))

        for i in range(len(self.soil_mats)):
            if hasattr(self.soil_mats[i], 'update_to_nonlinear'):
                self.soil_mats[i].update_to_nonlinear()
        for ele in self.eles:
            mat = ele.mat
            if hasattr(mat, 'set_nu'):
                mat.set_nu(mat.dynamic_poissons_ratio, eles=[ele])
        o3.analyze(self.osi, 40, 500.)

        # reset time and analysis
        o3.wipe_analysis(self.osi)

    def get_nearest_node_layer_at_depth(self, depth):
        # Convert to positive since node depths go downwards
        return int(np.round(np.interp(depth, -self.node_depths, np.arange(len(self.node_depths)))))

    def get_nearest_ele_layer_at_depth(self, depth):
        # Convert to positive since ele depths go downwards
        return int(np.round(np.interp(depth, -self.ele_depths, np.arange(len(self.ele_depths)))))

    def apply_loads(self, ray_freqs=(0.5, 10), xi=0.03):
        o3.set_time(self.osi, 0.0)

        # Define the dynamic analysis
        o3.constraints.Transformation(self.osi)
        o3.test.NormDispIncr(self.osi, tol=1.0e-4, max_iter=30, p_flag=0)
        # o3.test_check.EnergyIncr(self.osi, tol=1.0e-6, max_iter=30)
        o3.algorithm.Newton(self.osi)
        o3.system.SparseGeneral(self.osi)
        o3.numberer.RCM(self.osi)
        o3.integrator.Newmark(self.osi, gamma=0.5, beta=0.25)
        o3.analysis.Transient(self.osi)
        omega_1 = 2 * np.pi * ray_freqs[0]
        omega_2 = 2 * np.pi * ray_freqs[1]
        a0 = 2 * xi * omega_1 * omega_2 / (omega_1 + omega_2)
        a1 = 2 * xi / (omega_1 + omega_2)
        o3.rayleigh.Rayleigh(self.osi, a0, a1, 0, 0)

        static_time = 500
        print('time: ', o3.get_time(self.osi))
        # Add static stress bias
        time_series = o3.time_series.Path(self.osi, time=[0, static_time / 2, static_time, 1e3], values=[0, 0.5, 1, 1],
                                          use_last=True)
        o3.pattern.Plain(self.osi, time_series)
        net_hload = 0
        for i in range(len(self.sp.hloads)):
            pload = self.sp.hloads[i].p_x
            y = self.sp.hloads[i].y
            ind = self.get_nearest_node_layer_at_depth(y)
            print(i, y, ind)
            if self.sp.loads_are_stresses:
                pload *= self.ele_width
            o3.Load(self.osi, self.sn[ind][0], [pload, 0])
            net_hload += pload
        if self.base_imp >= 0:
            o3.Load(self.osi, self.sn[-1][0], [-net_hload, 0])

        static_dt = 0.1
        o3.analyze(self.osi, int(static_time / static_dt), static_dt)
        o3.load_constant(self.osi, time=0)

    def execute_dynamic(self, asig, analysis_dt=0.001, ray_freqs=(0.5, 10), xi=0.03, analysis_time=None,
                        outs=None, rec_dt=None, playback_dt=None, playback=True):
        self.rec_dt = rec_dt
        self.playback_dt = playback_dt
        if rec_dt is None:
            self.rec_dt = asig.dt
        if playback_dt is None:
            self.playback_dt = asig.dt
        if analysis_time is None:
            analysis_time = asig.time[-1]
        o3.set_time(self.osi, 0.0)

        # Define the dynamic analysis
        o3.constraints.Transformation(self.osi)
        o3.test.NormDispIncr(self.osi, tol=1.0e-4, max_iter=30, p_flag=0)
        # o3.test_check.EnergyIncr(self.osi, tol=1.0e-6, max_iter=30)
        o3.algorithm.Newton(self.osi)
        o3.system.SparseGeneral(self.osi)
        o3.numberer.RCM(self.osi)
        o3.integrator.Newmark(self.osi, gamma=0.5, beta=0.25)
        o3.analysis.Transient(self.osi)
        # Rayleigh damping parameters
        omega_1 = 2 * np.pi * ray_freqs[0]
        omega_2 = 2 * np.pi * ray_freqs[1]
        a0 = 2 * xi * omega_1 * omega_2 / (omega_1 + omega_2)
        a1 = 2 * xi / (omega_1 + omega_2)
        o3.rayleigh.Rayleigh(self.osi, a0, a1, 0, 0)

        init_time = o3.get_time(self.osi)
        if playback:
            self.o3res.dynamic = True
            self.o3res.start_recorders(self.osi, dt=self.playback_dt)
        else:
            self.o3res.dynamic = False
        self.o3sra_outs = O3SRAOutputs()
        self.o3sra_outs.start_recorders(self.osi, outs, self.sn, self.eles, rec_dt=self.rec_dt)

        # Define the dynamic input motion
        if self.base_imp < 0:  # fixed base
            acc_series = o3.time_series.Path(self.osi, dt=asig.dt, values=asig.values)
            o3.pattern.UniformExcitation(self.osi, dir=o3.cc.X, accel_series=acc_series)
        else:
            ts_obj = o3.time_series.Path(self.osi, dt=asig.dt, values=asig.velocity * 1, factor=self.c_base)
            o3.pattern.Plain(self.osi, ts_obj)
            o3.Load(self.osi, self.sn[-1][0], [1., 0.])
        if self.state == 3:
            o3.extensions.to_py_file(self.osi, self.opfile)
        # Run the dynamic motion
        o3.record(self.osi)
        while o3.get_time(self.osi) - init_time < analysis_time:
            if o3.analyze(self.osi, 1, analysis_dt):
                print('failed')
                if o3.analyze(self.osi, 10, analysis_dt / 10):
                    break
        o3.wipe(self.osi)
        self.out_dict = self.o3sra_outs.results_to_dict()

        if self.cache_path:
            self.o3sra_outs.cache_path = self.cache_path
            self.o3sra_outs.results_to_files()
            self.o3res.save_to_cache()


def run_sra(sp, asig, ray_freqs=(0.5, 10), xi=0.03, analysis_dt=0.001, dy=0.5, analysis_time=None, outs=None,
                  base_imp=0, k0=0.5, cache_path=None, opfile=None, playback=False, rec_dt=None):
    """

    Parameters
    ----------
    sp
    asig
    ray_freqs
    xi
    analysis_dt
    dy
    analysis_time
    outs
    base_imp: float
        If positive then use as impedence at base of model,
        If zero then use last soil layer
        If negative then use fixed base
    k0
    cache_path
    opfile
    playback

    Returns
    -------

    """
    sra_1d = SRA1D(sp, dy=dy, k0=k0, base_imp=base_imp, cache_path=cache_path, opfile=opfile)
    sra_1d.build_model()
    sra_1d.execute_static()
    if hasattr(sra_1d.sp, 'hloads'):
        sra_1d.apply_loads()
    sra_1d.execute_dynamic(asig, analysis_dt=analysis_dt, ray_freqs=ray_freqs, xi=xi, analysis_time=analysis_time,
                           outs=outs, playback=playback, playback_dt=0.01, rec_dt=rec_dt)
    return sra_1d



def site_response(sp, asig, freqs=(0.5, 10), xi=0.03, analysis_dt=0.001, dy=0.5, analysis_time=None, outs=None,
                  rec_dt=None, base_imp=0, cache_path=None, opfile=None, playback=False):
    """
    Run seismic analysis of a soil profile - example based on:
    http://opensees.berkeley.edu/wiki/index.php/Site_Response_Analysis_of_a_Layered_Soil_Column_(Total_Stress_Analysis)

    Parameters
    ----------
    sp: sfsimodels.SoilProfile object
        A soil profile
    asig: eqsig.AccSignal object
        An acceleration signal
    base_imp: float
        If positive then use as impedence at base of model,
        If zero then use last soil layer
        If negative then use fixed base

    Returns
    -------

    """
    if analysis_time is None:
        analysis_time = asig.time[-1]
    if outs is None:
        outs = {'ACCX': 'all'}  # Export the horizontal acceleration at the surface
    if rec_dt is None:
        rec_dt = analysis_dt
    else:
        raise ValueError('This is causing an error')
    state = 0
    if opfile:
        state = 3
    osi = o3.OpenSeesInstance(ndm=2, ndf=2, state=state)
    assert isinstance(sp, sm.SoilProfile)
    sp.gen_split(props=['shear_vel', 'unit_mass'], target=dy)
    thicknesses = sp.split["thickness"]
    n_node_rows = len(thicknesses) + 1
    node_depths = np.cumsum(sp.split["thickness"])
    node_depths = np.insert(node_depths, 0, 0)
    ele_depths = (node_depths[1:] + node_depths[:-1]) / 2
    unit_masses = sp.split["unit_mass"] / 1e3

    grav = 9.81
    # Rayleigh damping parameters
    omega_1 = 2 * np.pi * freqs[0]
    omega_2 = 2 * np.pi * freqs[1]
    a0 = 2 * xi * omega_1 * omega_2 / (omega_1 + omega_2)
    a1 = 2 * xi / (omega_1 + omega_2)

    k0 = 0.5
    pois = k0 / (1 + k0)

    ele_width = 3 * min(thicknesses)

    # Define nodes and set boundary conditions for simple shear deformation
    # Start at top and build down?
    nx = 1
    sn = []
    # sn = [[o3.node.Node(self.osi, ele_width * j, 0) for j in range(nx + 1)]]
    for i in range(0, n_node_rows):
        # Establish left and right nodes
        sn.append([o3.node.Node(osi, ele_width * j, -node_depths[i]) for j in range(nx + 1)])
        # set x and y dofs equal for left and right nodes
        o3.EqualDOF(osi, sn[i][0], sn[i][-1], [o3.cc.X, o3.cc.Y])  # if i != n_node_rows - 1:
    sn = np.array(sn)

    if base_imp < 0:  # Fixed base
        o3.Fix2DOFMulti(osi, sn[-1], o3.cc.FIXED, o3.cc.FIXED)  # Fix base nodes
    else:
        # Fix base nodes
        o3.Fix2DOFMulti(osi, sn[-1], o3.cc.FREE, o3.cc.FIXED)

        # Define dashpot nodes
        dashpot_node_l = o3.node.Node(osi, 0, -node_depths[-1])
        dashpot_node_2 = o3.node.Node(osi, 0, -node_depths[-1])
        o3.Fix2DOF(osi, dashpot_node_l, o3.cc.FIXED, o3.cc.FIXED)
        o3.Fix2DOF(osi, dashpot_node_2, o3.cc.FREE, o3.cc.FIXED)

        # define equal DOF for dashpot and soil base nodes
        o3.EqualDOF(osi, sn[-1][0], sn[-1][1], [o3.cc.X])
        o3.EqualDOF(osi, sn[-1][0], dashpot_node_2, [o3.cc.X])

    # define materials
    ele_thick = 1.0  # m
    soil_mats = []
    strains = np.logspace(-6, -0.5, 16)
    prev_args = []
    prev_kwargs = {}
    prev_sl_class = None
    eles = []
    for i in range(len(thicknesses)):
        y_depth = ele_depths[i]

        sl_id = sp.get_layer_index_by_depth(y_depth)
        sl = sp.layer(sl_id)

        app2mod = {}
        if y_depth > sp.gwl:
            umass = sl.unit_sat_mass / 1e3  # TODO: work out how to run in Pa, N, m, s
        else:
            umass = sl.unit_dry_mass / 1e3
        overrides = {'nu': pois, 'p_atm': 101,
                     'rho': umass,
                     'unit_moist_mass': umass,
                     'nd': 2.0,
                     # 'n_surf': 25
                     }
        # Define material
        if not hasattr(sl, 'o3_type'):
            sl.o3_type = sl.type
        if sl.o3_type == 'pm4sand':
            sl_class = o3.nd_material.PM4Sand
            # overrides = {'nu': pois, 'p_atm': 101, 'unit_moist_mass': umass}
            app2mod = sl.app2mod
        elif sl.o3_type == 'sdmodel':
            sl_class = o3.nd_material.StressDensity
            # overrides = {'nu': pois, 'p_atm': 101, 'unit_moist_mass': umass}
            app2mod = sl.app2mod
        elif sl.o3_type in ['pimy', 'pdmy', 'pdmy02']:
            if hasattr(sl, 'get_g_mod_at_m_eff_stress'):
                if hasattr(sl, 'g_mod_p0') and sl.g_mod_p0 != 0.0:
                    v_eff = sp.get_v_eff_stress_at_depth(y_depth)
                    k0 = sl.poissons_ratio / (1 - sl.poissons_ratio)
                    m_eff = v_eff * (1 + 2 * k0) / 3
                    p = m_eff  # Pa
                    overrides['d'] = 0.0
                else:
                    p = 101.0e3  # Pa
                    overrides['d'] = sl.a
                g_mod_r = sl.get_g_mod_at_m_eff_stress(p) / 1e3
            else:
                p = 101.0e3  # Pa
                overrides['d'] = 0.0
                g_mod_r = sl.g_mod / 1e3

            b_mod = 2 * g_mod_r * (1 + sl.poissons_ratio) / (3 * (1 - 2 * sl.poissons_ratio))
            overrides['p_ref'] = p / 1e3
            overrides['g_mod_ref'] = g_mod_r
            overrides['bulk_mod_ref'] = b_mod
            if sl.o3_type == 'pimy':
                overrides['cohesion'] = sl.cohesion / 1e3
                sl_class = o3.nd_material.PressureIndependMultiYield
            elif sl.o3_type == 'pdmy':
                sl_class = o3.nd_material.PressureDependMultiYield
            elif sl.o3_type == 'pdmy02':
                sl_class = o3.nd_material.PressureDependMultiYield02
                app2mod['e_init'] = 'e_curr'
        else:
            g_mod = sp.split['shear_vel'][i] ** 2 / unit_masses[i]
            sl_class = o3.nd_material.ElasticIsotropic
            sl.e_mod = 2 * g_mod * (1 - sl.poissons_ratio)
            overrides['nu'] = sl.poissons_ratio
            app2mod['rho'] = 'unit_moist_mass'
        args, kwargs = o3.extensions.get_o3_kwargs_from_obj(sl, sl_class, custom=app2mod, overrides=overrides)

        if o3.extensions.has_o3_model_changed(sl_class, prev_sl_class, args, prev_args, kwargs, prev_kwargs):
            mat = sl_class(osi, *args, **kwargs)
            prev_sl_class = sl_class
            prev_args = copy.deepcopy(args)
            prev_kwargs = copy.deepcopy(kwargs)
            mat.dynamic_poissons_ratio = sl.poissons_ratio
            soil_mats.append(mat)

        # def element
        for xx in range(nx):
            nodes = [sn[i+1][xx], sn[i+1][xx + 1], sn[i][xx + 1], sn[i][xx]]  # anti-clockwise
            #eles.append(o3.element.Quad(osi, nodes, ele_thick, o3.cc.PLANE_STRAIN, mat, b2=-grav * unit_masses[i]))
            eles.append(o3.element.SSPquad(osi, nodes, mat, o3.cc.PLANE_STRAIN, ele_thick, 0.0, -grav * unit_masses[i]))

    if base_imp >= 0:
        # define material and element for viscous dampers
        if base_imp == 0:
            sl = sp.get_soil_at_depth(sp.height)
            base_imp = sl.unit_dry_mass * sp.get_shear_vel_at_depth(sp.height)
        c_base = ele_width * base_imp / 1e3
        dashpot_mat = o3.uniaxial_material.Viscous(osi, c_base, alpha=1.)
        o3.element.ZeroLength(osi, [dashpot_node_l, dashpot_node_2], mats=[dashpot_mat], dirs=[o3.cc.DOF2D_X])

    coords = o3.get_all_node_coords(osi)
    ele_node_tags = o3.get_all_ele_node_tags_as_dict(osi)

    # Static analysis
    o3.constraints.Transformation(osi)
    o3.test.NormDispIncr(osi, tol=1.0e-5, max_iter=30, p_flag=0)
    o3.algorithm.Newton(osi)
    o3.numberer.RCM(osi)
    o3.system.ProfileSPD(osi)
    o3.integrator.Newmark(osi, gamma=0.5, beta=0.25)
    o3.analysis.Transient(osi)
    o3.analyze(osi, 10, 500.)

    for i in range(len(soil_mats)):
        if hasattr(soil_mats[i], 'update_to_nonlinear'):
            soil_mats[i].update_to_nonlinear()
    for ele in eles:
        mat = ele.mat
        if hasattr(mat, 'set_nu'):
            mat.set_nu(mat.dynamic_poissons_ratio, eles=[ele])
    o3.analyze(osi, 40, 500.)

    # reset time and analysis
    o3.wipe_analysis(osi)
    o3.set_time(osi, 0.0)
    if playback:
        all_node_xdisp_rec = o3.recorder.NodesToArrayCache(osi, 'all', [o3.cc.DOF2D_X], 'disp', nsd=4)
        all_node_ydisp_rec = o3.recorder.NodesToArrayCache(osi, 'all', [o3.cc.DOF2D_Y], 'disp', nsd=4)

    if hasattr(sp, 'hloads'):
        # Define the dynamic analysis
        o3.constraints.Transformation(osi)
        o3.test.NormDispIncr(osi, tol=1.0e-4, max_iter=30, p_flag=0)
        # o3.test_check.EnergyIncr(osi, tol=1.0e-6, max_iter=30)
        o3.algorithm.Newton(osi)
        o3.system.SparseGeneral(osi)
        o3.numberer.RCM(osi)
        o3.integrator.Newmark(osi, gamma=0.5, beta=0.25)
        o3.analysis.Transient(osi)
        # o3.rayleigh.Rayleigh(osi, a0, a1, 0, 0)
        pload = sp.hloads[0].p_x
        static_time = 100
        print('time: ', o3.get_time(osi))
        # Add static stress bias
        time_series = o3.time_series.Path(osi, time=[0, static_time / 2, static_time, 1e3], values=[0, 0.5, 1, 1], use_last=1)
        o3.pattern.Plain(osi, time_series)
        o3.Load(osi, sn[0][0], [pload * ele_width, 0])
        o3.Load(osi, sn[9][0], [-pload * ele_width, 0])
        if base_imp >= 0:
            o3.Load(osi, sn[-1][0], [-pload, 0])

        static_dt = 0.1
        o3.analyze(osi, int(static_time / static_dt) * 1.5, static_dt)
        o3.load_constant(osi, time=0)

        o3.wipe_analysis(osi)
    o3.set_time(osi, 0.0)  # TODO:
    # Define the dynamic analysis
    o3.constraints.Transformation(osi)
    o3.test.NormDispIncr(osi, tol=1.0e-4, max_iter=30, p_flag=0)
    # o3.test_check.EnergyIncr(osi, tol=1.0e-6, max_iter=30)
    o3.algorithm.Newton(osi)
    o3.system.SparseGeneral(osi)
    o3.numberer.RCM(osi)
    o3.integrator.Newmark(osi, gamma=0.5, beta=0.25)
    o3.analysis.Transient(osi)
    o3.rayleigh.Rayleigh(osi, a0, a1, 0, 0)

    init_time = o3.get_time(osi)
    o3sra_outs = O3SRAOutputs()
    o3sra_outs.start_recorders(osi, outs, sn, eles, rec_dt=rec_dt)

    # Define the dynamic input motion
    if base_imp < 0:  # fixed base
        acc_series = o3.time_series.Path(osi, dt=asig.dt, values=asig.values)
        o3.pattern.UniformExcitation(osi, dir=o3.cc.X, accel_series=acc_series)
    else:
        ts_obj = o3.time_series.Path(osi, dt=asig.dt, values=asig.velocity * 1, factor=c_base)
        o3.pattern.Plain(osi, ts_obj)
        o3.Load(osi, sn[-1][0], [1., 0.])
    if state == 3:
        o3.extensions.to_py_file(osi, opfile)
    # Run the dynamic motion
    o3.record(osi)
    while o3.get_time(osi) - init_time < analysis_time:
        if o3.analyze(osi, 1, analysis_dt):
            print('failed')
            break
    o3.wipe(osi)
    out_dict = o3sra_outs.results_to_dict()

    if cache_path:
        o3sra_outs.cache_path = cache_path
        o3sra_outs.results_to_files()
        o3res = o3.results.Results2D()
        o3res.cache_path = cache_path
        o3res.coords = coords
        o3res.ele2node_tags = ele_node_tags
        if playback:
            o3res.x_disp = all_node_xdisp_rec.collect()
            o3res.y_disp = all_node_ydisp_rec.collect()
        o3res.save_to_cache()

    return out_dict


def site_response_w_pysra(soil_profile, asig, odepths):
    print('site_response_w_pysra -> deprecated: use liquepy.sra.run_pysra')
    import liquepy as lq
    import pysra
    pysra_profile = lq.sra.sm_profile_to_pysra(soil_profile, d_inc=[0.5] * soil_profile.n_layers)
    # Should be input in g
    pysra_m = pysra.motion.TimeSeriesMotion(asig.label, None, time_step=asig.dt, accels=-asig.values / 9.8)

    calc = pysra.propagation.EquivalentLinearCalculator()

    od = {'ACCX': [], 'STRS': [], 'TAU': []}
    outs = []
    for i, depth in enumerate(odepths):
        od['ACCX'].append(len(outs))
        outs.append(pysra.output.AccelerationTSOutput(pysra.output.OutputLocation('within', depth=depth)))
        od['STRS'].append(len(outs))
        outs.append(pysra.output.StrainTSOutput(pysra.output.OutputLocation('within', depth=depth), in_percent=False))
        od['TAU'].append(len(outs))
        outs.append(pysra.output.StressTSOutput(pysra.output.OutputLocation('within', depth=depth),
                                                normalized=False))
    outputs = pysra.output.OutputCollection(outs)
    calc(pysra_m, pysra_profile, pysra_profile.location('outcrop', depth=soil_profile.height))
    outputs(calc)

    out_series = {}
    for mtype in od:
        out_series[mtype] = []
        for i in range(len(od[mtype])):
            out_series[mtype].append(outputs[od[mtype][i]].values[:asig.npts])
        out_series[mtype] = np.array(out_series[mtype])
        if mtype == 'ACCX':
            out_series[mtype] *= 9.8
    return out_series



import numpy as np


def set_params_from_op_pimy_model(sl, p_ref=100.0e3, hyp=True):
    # Octahedral shear stress
    tau_f = (2 * np.sqrt(2.) * np.sin(sl.phi_r)) / (3 - np.sin(sl.phi_r)) * p_ref + 2 * np.sqrt(2.) / 3 * sl.cohesion
    if hasattr(sl, 'get_g_mod_at_m_eff_stress'):
        g_mod_r = sl.get_g_mod_at_m_eff_stress(p_ref)
        if hasattr(sl, 'g_mod_p0'):
            assert sl.g_mod_p0 == 0.0
        d = sl.a
    else:
        g_mod_r = sl.g_mod
        d = 0.0
    strain_r = sl.peak_strain * tau_f / (g_mod_r * sl.peak_strain - tau_f)
    sdf = (p_ref / p_ref) ** d
    if hyp:  # hyperbolic model parameters
        sl.strain_curvature = 1.0
        sl.xi_min = 0.01
        dss_eq = 1.  # np.sqrt(3. / 2)  # correct to direct simple shear equivalent
        sl.strain_ref = strain_r / sdf / dss_eq
        sl.sra_type = "hyperbolic"
        sl.inputs += ['strain_curvature', 'xi_min', 'sra_type', 'strain_ref']

    sl.p_ref = p_ref
    sl.g_mod_ref = g_mod_r
    b_mod = 2 * g_mod_r * (1 + sl.poissons_ratio) / (3 * (1 - 2 * sl.poissons_ratio))
    sl.bulk_mod_ref = b_mod
    sl.inputs += ['p_ref', 'g_mod_ref', 'bulk_mod_ref']


def set_hyp_params_from_op_pimy_or_pdmy_model(sl):
    # Octahedral shear stress
    tau_f = (2 * np.sqrt(2.) * np.sin(sl.phi_r)) / (3 - np.sin(sl.phi_r)) * sl.p_ref + 2 * np.sqrt(2.) / 3 * sl.cohesion
    if hasattr(sl, 'get_g_mod_at_m_eff_stress'):
        g_mod_r = sl.get_g_mod_at_m_eff_stress(sl.p_ref)
        if hasattr(sl, 'g_mod_p0'):
            assert sl.g_mod_p0 == 0.0
        d = sl.a
    else:
        g_mod_r = sl.g_mod
        d = 0.0
    print('tau_f: ', tau_f)
    print('cohesion: ', sl.cohesion)
    strain_r = sl.peak_strain * tau_f / (g_mod_r * sl.peak_strain - tau_f)
    sdf = (sl.p_ref / sl.p_ref) ** d
    sl.strain_curvature = 1.0
    sl.xi_min = 0.01
    dss_eq = 1.  # np.sqrt(3. / 2)  # correct to direct simple shear equivalent
    sl.strain_ref = strain_r / sdf / dss_eq
    sl.sra_type = "hyperbolic"
    sl.inputs += ['strain_curvature', 'xi_min', 'sra_type', 'strain_ref']


def calc_backbone_op_pimy_model(sl, strains, esig_v0=100., ndm=2):
    k0 = sl.poissons_ratio / (1. - sl.poissons_ratio)
    p_eff = (esig_v0 * (1 + 2 * k0) / 3)
    # Octahedral shear stress
    tau_f = (2 * np.sqrt(2.) * np.sin(sl.phi_r)) / (3 - np.sin(sl.phi_r)) * p_eff + 2 * np.sqrt(2.) / 3 * sl.cohesion
    tau_f_ref = (2 * np.sqrt(2.) * np.sin(sl.phi_r)) / (3 - np.sin(sl.phi_r)) * sl.p_ref + 2 * np.sqrt(2.) / 3 * sl.cohesion
    if hasattr(sl, 'get_g_mod_at_v_eff_stress'):
        g_init = sl.get_g_mod_at_v_eff_stress(esig_v0)

        if sl.phi == 0.0:
            d = 0.
        else:
            d = sl.a
        g_mod_r = sl.g0_mod * (p_eff / sl.p_ref) ** d
    else:
        g_init = sl.g_mod
        g_mod_r = sl.g_mod
        d = 0.0

    dss_eq = np.sqrt(3. / 2)  # correct to direct simple shear equivalent
    strain_r = sl.peak_strain * tau_f_ref / (g_mod_r * sl.peak_strain - tau_f_ref) * dss_eq
    tau_back = g_init * strains / ((1 + strains / strain_r) * (sl.p_ref / p_eff) ** d)
    return tau_back